/*
	class to implement Adaptive Huffman Encoding
	derives from base Compress to inherit the bitwise i/o
	routines...
*/
#ifndef _COMPRESSHUF_H
	#define _COMPRESSHUF_H
#ifndef _COMPRESS_H
	#include "compress.h"
#endif
#ifndef _MSTRING_H
	#include "mstring.h"
#endif
#ifndef _INC_STDIO
	#include <stdio.h>
#endif
#ifndef __IOSTREAM_H
	#include <iostream>
#endif
#ifndef __TREE_H
	#include "tree.h"
#endif

using namespace std;

//some global definitions...
#define END_OF_STREAM     256
#define ESCAPE            257
#define ROOT_NODE         0
#define MAX_WEIGHT        0x8000
#define SYMBOL_COUNT      258
#define NODE_TABLE_COUNT  ( ( SYMBOL_COUNT * 2 ) - 1 )

class CompressHuff	: public Compress {
public:
	  CompressHuff ();
	~ CompressHuff ();
	void CompressFile( String &input, String &output );
	void ExpandFile( String &input, String &output );

protected:
	void InitializeTree();
	void EncodeSymbol( unsigned int c, bit_file &output );
	int  DecodeSymbol( bit_file &input );
	void UpdateModel( int c );
	void RebuildTree();
	void swap_nodes( int i, int j );
	void add_new_node( int c );
    /*
     * This data structure is all that is needed to maintain an adaptive
     * Huffman tree for both encoding and decoding.  The leaf array is a
     * set of indices into the nodes that indicate which node is the
     * parent of a symbol.  For example, to encode 'A', we would find the
     * leaf node by way of leaf[ 'A' ].  The next_free_node index is used
     * to tell which node is the next one in the array that can be used.
     * Since nodes are allocated when characters are read in for the first
     * time, this pointer keeps track of where we are in the node array.
     * Finally, the array of nodes is the actual Huffman tree.  The child
     * index is either an index pointing to a pair of children, or an
     * actual symbol value, depending on whether 'child_is_leaf' is true
     * or false.
     */
	/*
    struct Tree {
        int leaf[ SYMBOL_COUNT ];
        int next_free_node;
        struct node {
            unsigned int weight;
            int parent;
            int child_is_leaf;
            int child;
        } nodes[ NODE_TABLE_COUNT ];
		
	};*/

    Tree m_tree;
};
#endif