#include <stdio.h>

#include "mstring.h"
#include "tree.h"
#include "comphuff.h"

/*
#ifdef	WIN32
	#ifndef _CONSOLE
		#include <windows.h>
	#endif
#endif
*/

CompressHuff::CompressHuff()
:Compress()
{
	typeChar = '%';
}

CompressHuff::~CompressHuff()
{
	//nothing to destruct
}

void CompressHuff::CompressFile( String &input, String &output )
{
	// HOWARD
	// changed from original int c to unsigned char to handle
	// arbitrary bit file.
	// int c;
    unsigned char c;
	bit_file	*in;
	bit_file	*out;
    
	InitializeTree();
	in  = OpenInputBitFile( input );
	out = OpenOutputBitFile( output );
    
	/*
	while (!in->IsEOF()) {
        c = InputStd( in );

		// TODO: HOWARD
		// I put this test here because without it 
		// it will read past the EOF
		//if (c < 0) 
		//	break;


        // EncodeSymbol( c, *out );
        // UpdateModel( c );
    }
	*/

	while(in != NULL){
		c = InputStd(in);
		
		// HOWARD
		// since the read operation happens in InputStd(in)
		// we have to test IsEOF after the read operation
		if(in->IsEOF())
			break;

        EncodeSymbol( c, *out );
        UpdateModel( c );
	}

	// build the code book here
    EncodeSymbol( END_OF_STREAM, *out );
	CloseOutputBitFile( out );
	CloseInputBitFile( in );
}

void CompressHuff::ExpandFile( String &input, String &output )
{
    int			c;
	bit_file    *in;
	bit_file    *out;
    
	InitializeTree();
	in  = OpenInputBitFile( input );
	out = OpenOutputBitFile( output );
    while ( ( c = DecodeSymbol( *in ) ) != END_OF_STREAM ) {
        if ( OutputStd( c, out ) == EOF )
            throw( "CompressHuff::Error writing character" );
        UpdateModel( c );
    }
	CloseOutputBitFile( out );
	CloseInputBitFile( in );
}

void CompressHuff::InitializeTree()
{
    int i;

    m_tree.nodes[ ROOT_NODE ].child             = ROOT_NODE + 1;
    m_tree.nodes[ ROOT_NODE ].child_is_leaf     = 0;
    m_tree.nodes[ ROOT_NODE ].weight            = 2;
    m_tree.nodes[ ROOT_NODE ].parent            = -1;

    m_tree.nodes[ ROOT_NODE + 1 ].child         = END_OF_STREAM;
    m_tree.nodes[ ROOT_NODE + 1 ].child_is_leaf = 1;
    m_tree.nodes[ ROOT_NODE + 1 ].weight        = 1;
    m_tree.nodes[ ROOT_NODE + 1 ].parent        = ROOT_NODE;
    m_tree.leaf[ END_OF_STREAM ]                = ROOT_NODE + 1;

    m_tree.nodes[ ROOT_NODE + 2 ].child         = ESCAPE;
    m_tree.nodes[ ROOT_NODE + 2 ].child_is_leaf = 1;
    m_tree.nodes[ ROOT_NODE + 2 ].weight        = 1;
    m_tree.nodes[ ROOT_NODE + 2 ].parent        = ROOT_NODE;
    m_tree.leaf[ ESCAPE ]                       = ROOT_NODE + 2;

    m_tree.next_free_node                       = ROOT_NODE + 3;

    for ( i = 0 ; i < END_OF_STREAM ; i++ )
        m_tree.leaf[ i ] = -1;
}

void CompressHuff::EncodeSymbol( unsigned int c, bit_file &output )
{
    unsigned long code;
    unsigned long current_bit;
    int code_size;
    int current_node;

    code = 0;
    current_bit = 1;
    code_size = 0;
    current_node = m_tree.leaf[ c ];
    if ( current_node == -1 )
        current_node = m_tree.leaf[ ESCAPE ];

	// build the code tree here and record the size
    while ( current_node != ROOT_NODE ) {
        if ( ( current_node & 1 ) == 0 )
            code |= current_bit;
        current_bit <<= 1;
        code_size++;
        current_node = m_tree.nodes[ current_node ].parent;
    };
	// store the code tree into the output file
    OutputBits( output, code, code_size );


    if ( m_tree.leaf[ c ] == -1 ) {
        OutputBits( output, (unsigned long) c, 8 );
        add_new_node( c );
    }
}
/*
 * Decoding symbols is easy.  We start at the root node, then go down
 * the tree until we reach a leaf.  At each node, we decide which
 * child to take based on the next input bit.  After getting to the
 * leaf, we check to see if we read in the ESCAPE code.  If we did,
 * it means that the next symbol is going to come through in the next
 * eight bits, unencoded.  If that is the case, we read it in here,
 * and add the new symbol to the table.
 */
int CompressHuff::DecodeSymbol( bit_file &input )
{
    int current_node;
    int c;

    current_node = ROOT_NODE;
    while ( !m_tree.nodes[ current_node ].child_is_leaf ) {
        current_node = m_tree.nodes[ current_node ].child;
        current_node += InputBit( input );
    }
    c = m_tree.nodes[ current_node ].child;

	// read in unencoded symbol
    if ( c == ESCAPE ) {
        c = (int) InputBits( input, 8 );
        add_new_node( c );
    }
    return( c );
}

/*
 * Adding a new node to the tree is pretty simple.  It is just a matter
 * of splitting the lightest-weight node in the tree, which is the highest
 * valued node.  We split it off into two new nodes, one of which is the
 * one being added to the tree.  We assign the new node a weight of 0,
 * so the tree doesn't have to be adjusted.  It will be updated later when
 * the normal update process occurs.  Note that this code assumes that
 * the lightest node has a leaf as a child.  If this is not the case,
 * the tree would be broken.
 */
void CompressHuff::add_new_node( int c )
{
    int lightest_node;
    int new_node;
    int zero_weight_node;

    lightest_node = m_tree.next_free_node - 1;
    new_node = m_tree.next_free_node;
    zero_weight_node = m_tree.next_free_node + 1;
    m_tree.next_free_node += 2;

    m_tree.nodes[ new_node ] = m_tree.nodes[ lightest_node ];
    m_tree.nodes[ new_node ].parent = lightest_node;
    m_tree.leaf[ m_tree.nodes[ new_node ].child ]	= new_node;

    m_tree.nodes[ lightest_node ].child				= new_node;
    m_tree.nodes[ lightest_node ].child_is_leaf		= 0;

    m_tree.nodes[ zero_weight_node ].child          = c;
    m_tree.nodes[ zero_weight_node ].child_is_leaf  = 1;
    m_tree.nodes[ zero_weight_node ].weight         = 0;
    m_tree.nodes[ zero_weight_node ].parent         = lightest_node;
    m_tree.leaf[ c ] = zero_weight_node;
}

/*
 * UpdateModel is called to increment the count for a given symbol.
 * After incrementing the symbol, this code has to work its way up
 * through the parent nodes, incrementing each one of them.  That is
 * the easy part.  The hard part is that after incrementing each
 * parent node, we have to check to see if it is now out of the proper
 * order.  If it is, it has to be moved up the tree into its proper
 * place.
 */
void CompressHuff::UpdateModel( int c )
{
    int current_node;
    int new_node;

    if ( m_tree.nodes[ ROOT_NODE].weight == MAX_WEIGHT )
        RebuildTree();
    current_node = m_tree.leaf[ c ];
    while ( current_node != -1 ) {
        m_tree.nodes[ current_node ].weight++;
        for ( new_node = current_node ; new_node > ROOT_NODE ; new_node-- )
            if ( m_tree.nodes[ new_node - 1 ].weight >=
                 m_tree.nodes[ current_node ].weight )
                break;
        if ( current_node != new_node ) {
            swap_nodes( current_node, new_node );
            current_node = new_node;
        }
        current_node = m_tree.nodes[ current_node ].parent;
    }
}

/*
 * Rebuilding the tree takes place when the counts have gone too
 * high.  From a simple point of view, rebuilding the tree just means that
 * we divide every count by two.  Unfortunately, due to truncation effects,
 * this means that the tree's shape might change.  Some nodes might move
 * up due to cumulative increases, while others may move down.
 */
void CompressHuff::RebuildTree( void )
{
    int i;
    int j;
    int k;
    unsigned int weight;

/*
 * To start rebuilding the table,  I collect all the leaves of the Huffman
 * tree and put them in the end of the tree.  While I am doing that, I
 * scale the counts down by a factor of 2.
 */
    j = m_tree.next_free_node - 1;
    for ( i = j ; i >= ROOT_NODE ; i-- ) {
        if ( m_tree.nodes[ i ].child_is_leaf ) {
            m_tree.nodes[ j ] = m_tree.nodes[ i ];
            m_tree.nodes[ j ].weight = ( m_tree.nodes[ j ].weight + 1 ) / 2;
            j--;
        }
    }

/*
 * At this point, j points to the first free node.  I now have all the
 * leaves defined, and need to start building the higher nodes on the
 * tree. I will start adding the new internal nodes at j.  Every time
 * I add a new internal node to the top of the tree, I have to check to
 * see where it really belongs in the tree.  It might stay at the top,
 * but there is a good chance I might have to move it back down.  If it
 * does have to go down, I use the memmove() function to scoot everyone
 * bigger up by one node.  Note that memmove() may have to be change
 * to memcpy() on some UNIX systems.  The parameters are unchanged, as
 * memmove and  memcpy have the same set of parameters.
 */
    for ( i = m_tree.next_free_node - 2 ; j >= ROOT_NODE ; i -= 2, j-- ) {
        k = i + 1;
        m_tree.nodes[ j ].weight = m_tree.nodes[ i ].weight +
                                   m_tree.nodes[ k ].weight;
        weight = m_tree.nodes[ j ].weight;
        m_tree.nodes[ j ].child_is_leaf = 0;
        for ( k = j + 1 ; weight < m_tree.nodes[ k ].weight ; k++ )
            ;
        k--;
        memmove( &(m_tree.nodes[ j ]), &(m_tree.nodes[ j + 1 ]),
			( k - j ) * sizeof( struct Tree::node ) );
        m_tree.nodes[ k ].weight = weight;
        m_tree.nodes[ k ].child = i;
        m_tree.nodes[ k ].child_is_leaf = 0;
    }
/*
 * The final step in tree reconstruction is to go through and set up
 * all of the leaf and parent members.  This can be safely done now
 * that every node is in its final position in the tree.
 */
    for ( i = m_tree.next_free_node - 1 ; i >= ROOT_NODE ; i-- ) {
        if ( m_tree.nodes[ i ].child_is_leaf ) {
            k = m_tree.nodes[ i ].child;
            m_tree.leaf[ k ] = i;
        } else {
            k = m_tree.nodes[ i ].child;
            m_tree.nodes[ k ].parent = m_tree.nodes[ k + 1 ].parent = i;
        }
    }
}

/*
 * Swapping nodes takes place when a node has grown too big for its
 * spot in the tree.  When swapping nodes i and j, we rearrange the
 * tree by exchanging the children under i with the children under j.
 */
void CompressHuff::swap_nodes( int i, int j )
{
    struct Tree::node temp;

    if ( m_tree.nodes[ i ].child_is_leaf )
         m_tree.leaf[ m_tree.nodes[ i ].child ] = j;
    else {
        m_tree.nodes[ m_tree.nodes[ i ].child ].parent = j;
        m_tree.nodes[ m_tree.nodes[ i ].child + 1 ].parent = j;
    }
    if ( m_tree.nodes[ j ].child_is_leaf )
         m_tree.leaf[ m_tree.nodes[ j ].child ] = i;
    else {
        m_tree.nodes[ m_tree.nodes[ j ].child ].parent = i;
        m_tree.nodes[ m_tree.nodes[ j ].child + 1 ].parent = i;
    }
    temp = m_tree.nodes[ i ];
    m_tree.nodes[ i ] = m_tree.nodes[ j ];
    m_tree.nodes[ i ].parent = temp.parent;
    temp.parent = m_tree.nodes[ j ].parent;
    m_tree.nodes[ j ] = temp;
}

