#include <stdio.h>
#include "mstring.h"
#include "complzw.h"


/*
#ifdef	WIN32
	#ifndef _CONSOLE
		#include <windows.h>
	#endif
#endif
*/


CompressLZW::CompressLZW()
: Compress()
{
	typeChar = '#';
	//insure everything is initially pointing to NULL
	for(int i = 0;i < TABLE_SIZE;i++)   {
		m_pDict[i] = NULL;
    }
}

CompressLZW::~CompressLZW()
{
	//do our cleanup...this is a LARGE data structure...
	for(int i =0;i < TABLE_SIZE;i++)    {
		if(m_pDict[i] != NULL)  {
			delete m_pDict[i];
        }
    }
}

/*
 * The compressor is short and simple.  It reads in new symbols one
 * at a time from the input file.  It then  checks to see if the
 * combination of the current symbol and the current code are already
 * defined in the dictionary.  If they are not, they are added to the
 * dictionary, and we start over with a new one symbol code.  If they
 * are, the code for the combination of the code and character becomes
 * our new code.  Note that in this enhanced version of LZW, the
 * encoder needs to check the codes for boundary conditions.
 */
void CompressLZW::CompressFile( String &input, String &output )
{
    int character;
    int string_code;
    unsigned int index;
	bit_file	*in;
	bit_file	*out;
    
    InitializeStorage();
    InitializeDictionary();

	in  = OpenInputBitFile( input );
	out = OpenOutputBitFile( output );
    
    string_code = in->GetBit();
    if ( in->IsEOF() )  {
        string_code = END_OF_STREAM;
    }
    while ( in->IsEOF() == 0 ) {
        character = in->GetBit();
        index = find_child_node( string_code, character );
        if ( m_pDict[index]->code_value != - 1 )    {
            string_code = m_pDict[index]->code_value;
        }
        else {
            m_pDict[index]->code_value = next_code++;
            m_pDict[index]->parent_code = string_code;
            m_pDict[index]->character = (char) character;
            OutputBits( *out, (unsigned long) string_code, current_code_bits );
            string_code = character;
            if ( next_code > MAX_CODE ) {
                OutputBits( *out,(unsigned long) FLUSH_CODE, current_code_bits );
                InitializeDictionary();
            } else if ( next_code > next_bump_code ) {
                OutputBits( *out, (unsigned long) BUMP_CODE, current_code_bits );
                current_code_bits++;
                next_bump_code <<= 1;
                next_bump_code |= 1;
#ifndef	WINVER	//not a WIN16 or WIN32 app
				putc( typeChar, stdout );
#else
	#ifndef _CONSOLE	//not a WIN32 Console app
				MessageBeep(MB_OK);
	#else
				putc( typeChar, stdout );
	#endif
#endif
            }
        }
    } 
    OutputBits( *out, (unsigned long) string_code, current_code_bits );
    OutputBits( *out, (unsigned long) END_OF_STREAM, current_code_bits);
	CloseOutputBitFile( out );
	CloseInputBitFile( in );
}

/*
 * The file expander operates much like the encoder.  It has to
 * read in codes, the convert the codes to a string of characters.
 * The only catch in the whole operation occurs when the encoder
 * encounters a CHAR+STRING+CHAR+STRING+CHAR sequence.  When this
 * occurs, the encoder outputs a code that is not presently defined
 * in the table.  This is handled as an exception.  All of the special
 * input codes are handled in various ways.
 */
void CompressLZW::ExpandFile( String &input, String &output )
{
    unsigned int new_code;
    unsigned int old_code;
    int character;
    unsigned int count;
	bit_file    *in;
	bit_file    *out;
    
	in  = OpenInputBitFile( input );
	out = OpenOutputBitFile( output );
    InitializeStorage();
    for ( ; ; ) {
        InitializeDictionary();
        old_code = (unsigned int) InputBits( *in, current_code_bits );
        if ( old_code == END_OF_STREAM )    {
        	CloseOutputBitFile( out );
        	CloseInputBitFile( in );
            return;
        }
        character = old_code;
        out->PutBit(old_code);
        for ( ; ; ) {
            new_code = (unsigned int) InputBits( *in, current_code_bits );
            if ( new_code == END_OF_STREAM )    {
            	CloseOutputBitFile( out );
            	CloseInputBitFile( in );
                return;
            }
            if ( new_code == FLUSH_CODE )   {
                break;
            }
            if ( new_code == BUMP_CODE ) {
                current_code_bits++;
#ifndef	WINVER	//not a WIN16 or WIN32 app
				putc( typeChar, stdout );
#else
	#ifndef _CONSOLE	//not a WIN32 Console app
				MessageBeep(MB_OK);
	#else
				putc( typeChar, stdout );
	#endif
#endif
                continue;
            }
            if ( new_code >= next_code ) {
                decode_stack[ 0 ] = (char) character;
                count = decode_string( 1, old_code );
            }
            else    {
                count = decode_string( 0, new_code );
            }
            character = decode_stack[ count - 1 ];
            while ( count > 0 ) {
                out->PutBit(decode_stack[ --count ]);
            }
            m_pDict[next_code]->parent_code = old_code;
            m_pDict[next_code]->character = (char) character;
            next_code++;
            old_code = new_code;
        }   
    }
	CloseOutputBitFile( out );
	CloseInputBitFile( in );
}

/*
 * This routine is used to initialize the dictionary, both when the
 * compressor or decompressor first starts up, and also when a flush
 * code comes in.  Note that even though the decompressor sets all
 * the code_value elements to UNUSED, it doesn't really need to.
 */
void CompressLZW::InitializeDictionary()
{
    unsigned int i;

    for ( i = 0 ; i < TABLE_SIZE ; i++ )    {
        m_pDict[i]->code_value = UNUSED;
    }
    next_code = FIRST_CODE;
#ifndef	WINVER	//not a WIN16 or WIN32 app
	putc( typeChar, stdout );
#else
	#ifndef _CONSOLE	//not a WIN32 Console app
	MessageBeep(MB_OK);
	#else
	putc( typeChar, stdout );
	#endif
#endif
    current_code_bits = 9;
    next_bump_code = 511;
}

/*
 * This routine allocates the dictionary.  The total size of the
 * dictionary is much larger than 64K, it can't be allocated as a single
 * object on all ports. 
 *
 */
void CompressLZW::InitializeStorage()
{
    unsigned int i;

    for ( i = 0 ; i < TABLE_SIZE ; i++ ) {
        m_pDict[ i ] =  new struct CompressLZW::dictionary;
        if ( m_pDict[ i ] == NULL )
            throw( "CompressLZW::Error allocating dictionary space" );
    }
}

/*
 * This hashing routine is responsible for finding the table location
 * for a string/character combination.  The table index is created
 * by using an exclusive OR combination of the prefix and character.
 * This code also has to check for collisions, and handles them by
 * jumping around in the table.
 */
unsigned int CompressLZW::find_child_node( int parent_code, int child_character )
{
    unsigned int index;
    unsigned int offset;

    index = ( child_character << ( BITS - 8 ) ) ^ parent_code;
    if ( index == 0 )
        offset = 1;
    else
        offset = TABLE_SIZE - index;
    if(index > TABLE_SIZE)  {
#ifndef	WINVER	//not a WIN16 or WIN32 app
		printf( "Bad index!!!" );
#else
	#ifndef _CONSOLE	//not a WIN32 Console app
		MessageBeep(MB_ICONERROR);
	#else
		printf( "Bad index!!!" );
	#endif
#endif
        return 0;
    }
    for ( ; ; ) {
        if ( m_pDict[index]->code_value == UNUSED )
            return( (unsigned int) index );
        if ( m_pDict[index]->parent_code == parent_code &&
             m_pDict[index]->character == (char) child_character )
            return( index );
        if ( (int) index >= offset )
            index -= offset;
        else
            index += TABLE_SIZE - offset;
    }
}

/*
 * This routine decodes a string from the dictionary, and stores it
 * in the decode_stack data structure.  It returns a count to the
 * calling program of how many characters were placed in the stack.
 */
unsigned int CompressLZW::decode_string( unsigned int count, unsigned int code )
{
    if(code > TABLE_SIZE){
#ifndef	WINVER	//not a WIN16 or WIN32 app
		printf( "Bad Code!!!" );
#else
	#ifndef _CONSOLE	//not a WIN32 Console app
		MessageBeep(MB_ICONERROR);
	#else
		printf( "Bad Code!!!" );
	#endif
#endif
        return 0;
    }
    while ( code > 255 ) {
        decode_stack[ count++ ] = m_pDict[code]->character;
        code = m_pDict[code]->parent_code;
    }
    decode_stack[ count++ ] = (char) code;
    return( count );
}


