/**********************************************************************
 *
 * This is the LZW compression class.
 * This version of the code expands the maximum code size
 * to 15 bits. It starts encoding with 9 bit codes, working
 * its way up in bit size only as necessary.  Finally, it flushes the
 * dictionary when done.
 *
 *	derives from base Compress to inherit the bitwise i/o                    
 *	routines...
 **********************************************************************/
#ifndef _COMPRESSLZW_H
	#define _COMPRESSLZW_H
#ifndef _COMPRESS_H
	#include "compress.h"
#endif
#ifndef _MSTRING_H
	#include "mstring.h"
#endif
#ifndef _INC_STDIO
	#include <stdio.h>
#endif
#ifndef __IOSTREAM_H
	#include <iostream>
#endif

using namespace std;

 /*
 * Constants used throughout the program.  BITS defines the maximum
 * number of bits that can be used in the output code.  TABLE_SIZE 
 * defines the size of the dictionary table.
 * The code defines should be self-explanatory.
 */

#define BITS                       15
#define MAX_CODE                   ( ( 1 << BITS ) - 1 )
#define TABLE_SIZE                 35023
#define END_OF_STREAM              256
#define BUMP_CODE                  257
#define FLUSH_CODE                 258
#define FIRST_CODE                 259
#define UNUSED                     -1

class CompressLZW : public Compress {

   /*
	* This data structure defines the dictionary.  Each entry in the dictionary
	* has a code value.  This is the code emitted by the compressor.  Each
	* code is actually made up of two pieces:  a parent_code, and a
	* character.  Code values of less than 256 are actually plain
	* text codes.
	*/

	struct dictionary {
		int code_value;
	    int parent_code;
	    char character;
	};

	struct dictionary *m_pDict[TABLE_SIZE];

/*
 * Other data structures.  The decode_stack is used to reverse
 * strings that come out of the tree during decoding.  next_code is the
 * next code to be added to the dictionary, both during compression and
 * decompression.  current_code_bits defines how many bits are currently
 * being used for output, and next_bump_code defines the code that will
 * trigger the next jump in word size.
 */

char decode_stack[ TABLE_SIZE ];
unsigned short int next_code;
int current_code_bits;
unsigned short int next_bump_code;

public:
	CompressLZW();
	~CompressLZW();
	void CompressFile( String &input, String &output );
	void ExpandFile( String &input, String &output );
protected:
	void InitializeDictionary(void);
	void InitializeStorage(void);
	unsigned int find_child_node( int parent_code, int child_character );
	unsigned int decode_string( unsigned int count, unsigned int code );
};

#endif
