/*
    class used to accomplish encryption and decryption. This is
    more roubst than PassWd() in that it applies a stronger
    encryption algorythm, (polyalphabetic - 512 byte key).
    Also this was designed with the idea of application to
    an entire file rather than to a single text buffer, although
    the capability to encode/decode a buffer at a time is also
    included. The algorythm employed here is the same one used in 
    WinCrypt - originally published by PcMag.
*/
#include <stdio.h>
#include "mstring.h"
#include "encrypt.h"

//Encrypt::Encrypt() : Pump()
Encrypt::Encrypt()
{
    //a default password
    npPassword = "DEADBEEF";
    Init = 0;
}

Encrypt::~Encrypt()
{
}

//.................................
//public functions.................
//.................................

/*
 *  encrypts or unencrypts a file.
 *
 *  Returns:
 *   -1 = Failed to allocate Memory for temp buffer
 *    0 = File was successfully encrypted or unencrypted
 *    1 = File was not found or could not be opened
 *    2 = Error reading file
 *    3 = Error writing file
 *
 *  Notes:
 *
 *    This function assumes that the key size and file I/O buffer size
 *    are the same. File data is read nKeySize bytes at a time, XORed with
 *    the corresponding characters in the encryption key, and written back
 *    to disk.
 */
int Encrypt::CryptFile(String &szInFile, String& szOutFile)
{
    FILE           *IFile;
    FILE           *OFile;
    unsigned int    nBytesToRead, nBytesRead, i;
    unsigned int    nBytesToWrite, nBytesWritten;
    unsigned long   dwBytesRemaining;
    char           *npFileData;

    if(!Init)
        InitEncrypt();

    npFileData = new char[npKey.getlen()];

    if(npFileData == NULL)
        return -1;
    //
    // Open the files for reading and writing.
    //
    if ((IFile = fopen(szInFile.c_str(), "rb")) == NULL)   {
        delete [] npFileData;
        return 1;
    }
    if ((OFile = fopen(szOutFile.c_str(), "wb")) == NULL)   {
        delete [] npFileData;
        fclose (IFile);
        return 1;
    }

    //
    // Determine the file size.         
    //
    fseek (IFile, 0l, SEEK_END);
    dwBytesRemaining = (unsigned long) ftell(IFile); 
    fseek (IFile, 0l, SEEK_SET);
    if(ftell(IFile) != 0l)  {
        delete [] npFileData;
        fclose (IFile);
        fclose (OFile);
        return 2;
    }

    //
    // Encrypt or unencrypt the file.
    //
    while (dwBytesRemaining) {
        nBytesToRead = __min((unsigned int)npKey.getlen(), dwBytesRemaining);
        nBytesRead = fread (npFileData, sizeof(char), nBytesToRead, IFile);

        if (nBytesRead != nBytesToRead) {
            fclose (IFile);
            fclose (OFile);
            delete [] npFileData;
            return 2;        
        }

        for (i=0; i<nBytesRead; i++)
            npFileData[i] ^= npKey[(int)i];

        nBytesToWrite = nBytesRead;
        nBytesWritten = fwrite (npFileData, sizeof(char), nBytesToWrite, OFile);
        if(nBytesWritten != nBytesToWrite) {
            fclose (IFile);
            fclose (OFile);
            delete [] npFileData;
            return 3;        
        }
        dwBytesRemaining -= (unsigned long) nBytesWritten;
    }

    //
    // Close the file and return 0.
    //
    fclose (IFile);
    fclose (OFile);
    delete [] npFileData;
    return 0;
}

/*
 *  encrypts or unencrypts a buffer
 *
 *  Returns:
 *   -1 = Failed to allocate Memory for temp. buffer
 *    0 = input buffer was successfully encrypted or unencrypted
 *
 */
int Encrypt::CryptBuffer(String &szBuffer)
{
    int            nBytesToRead, nPosition, i; 
    unsigned int   dwBytesRemaining;
    char          *npBuffData;

    if(!Init)
        InitEncrypt();

    npBuffData = new char[npKey.getlen()];

    if(npBuffData == NULL)
        return -1;

    dwBytesRemaining = (unsigned int) szBuffer.getlen(); 

    //
    // Encrypt or unencrypt the buffer
    //
    nPosition = 0;
    while (dwBytesRemaining) {
        nBytesToRead = __min((unsigned int)npKey.getlen(), dwBytesRemaining);
        memset(npBuffData,0,npKey.getlen());
        szBuffer.copy(npBuffData,nPosition,nBytesToRead);

        for (i=0; i<nBytesToRead; i++)
            npBuffData[i] ^= npKey[(int)i];
        
        for (i=nPosition; i<(nPosition + nBytesToRead); i++)
            szBuffer.set_at(i,npBuffData[i - nPosition]);

        dwBytesRemaining -= (unsigned int) nBytesToRead;
    }

    delete [] npBuffData;
    return 0;
}

//.................................
//protected functions..............
//.................................
/*
    Initializes the encryption engine. Creates the key from the 
    password. The password is used basically as the 'seed' from 
    which the key is grown.
*/
void Encrypt::InitEncrypt ()
{
    int     nPasswordLength, nCount, i, j;
    String  npTmp;
    char    c;

    //
    // Replicate the password enough times to fill the key.
    //
    nPasswordLength = npPassword.getlen();
    npKey.clear();
    npKey.allocate(KEYSIZE);
    npTmp.allocate(KEYSIZE);

    i = 0;
    while (i < KEYSIZE) {
        nCount = __min(nPasswordLength, KEYSIZE - i);
        j = 0;
        while (j < nCount)
            npKey.set_at(i++,npPassword[j++]);
    }

    //
    // Generate a series of "mutated" numbers from the key using a simple
    // (but non-reversible) one-step cellular automaton process, and XOR the
    // series into the key. 
    //
    Mutate (npKey, npTmp, (KEYSIZE/8), (KEYSIZE/8));
    i = 0;
    while (i < KEYSIZE) {
        c = npKey[i];
        c ^= npTmp[i];
        npKey.set_at(i++,c); 
    }

    //
    // Generate a series of random numbers and XOR the series into the key.
    //
    i = j = 0;
    while (i < nPasswordLength)
        j += (int) npPassword[i++];

    srand (j);

    i = 0;
    while (i < KEYSIZE) {
        c = npKey[i];
        c ^= (char) (rand () % 0x100);
        npKey.set_at(i++,c);
    }
}

/*
   generate a series of mutated keys... 
*/
void Encrypt::Mutate(String &npInputKey, String &npOutputKey, unsigned int nWidth, unsigned int nHeight)
{
    int             bCurrentCellAlive;
    unsigned int    nLastRow, nNextRow, nLastCol, nNextCol;
    unsigned int    i, j, nNeighbors;

    for (i=0; i<nHeight; i++) {
        nLastRow = (i == 0) ? nHeight - 1 : i - 1;
        nNextRow = (i == (nHeight - 1)) ? 0 : i + 1;

        for (j=0; j<nWidth; j++) {
            nLastCol = (j == 0) ? nWidth - 1 : j - 1;
            nNextCol = (j == (nWidth - 1)) ? 0 : j + 1;

            nNeighbors = 0;

            if (GetCell (npInputKey, nWidth, nLastCol, nLastRow))
                nNeighbors++;
            if (GetCell (npInputKey, nWidth, j, nLastRow))
                nNeighbors++;
            if (GetCell (npInputKey, nWidth, nNextCol, nLastRow))
                nNeighbors++;

            if (GetCell (npInputKey, nWidth, nLastCol, i))
                nNeighbors++;
            if (GetCell (npInputKey, nWidth, nNextCol, i))
                nNeighbors++;

            if (GetCell (npInputKey, nWidth, nLastCol, nNextRow))
                nNeighbors++;
            if (GetCell (npInputKey, nWidth, j, nNextRow))
                nNeighbors++;
            if (GetCell (npInputKey, nWidth, nNextCol, nNextRow))
                nNeighbors++;

            bCurrentCellAlive = GetCell (npInputKey, nWidth, j, i);

            if ((bCurrentCellAlive && ((nNeighbors == 2) ||
                (nNeighbors == 3))) || (!bCurrentCellAlive &&
                (nNeighbors == 3)))
                SetCell (npOutputKey, nWidth, j, i, 1);
            else
                SetCell (npOutputKey, nWidth, j, i, 0);
        }
    }
}

/*
 *  GetCell is called by Mutate to determine whether a specified cell is
 *  currently set to 0 or 1.
 *
 *  Input parameters:
 *
 *    npGrid = Pointer to grid
 *    nWidth = Grid width
 *    i = Cell's column number
 *    j = Cell's row number
 *
 *  Returns:
 *
 *    TRUE = Cell's value is 1
 *    FALSE = Cell's value is 0
 */
int Encrypt::GetCell (String &npGrid, unsigned int nWidth, unsigned int i, unsigned int j)
{
    unsigned int nByteNumber, nBitNumber;

    nBitNumber = ((j * nWidth) + i) & 7;
    nByteNumber = ((j * nWidth) + i) >> 3;
    return ((npGrid[(int)nByteNumber]) & (1 << nBitNumber)) ? 1 : 0;
}

/*
 *  SetCell is called by Mutate to set a cell value to 0 or 1. 
 *
 *  Input parameters:
 *
 *    npGrid = Pointer to grid
 *    i = Cell's column number
 *    j = Cell's row number
 *    bCellState = New cell value (TRUE=1, FALSE=0)
 *
 *  Returns:
 *
 *    Nothing
 */
void Encrypt::SetCell (String &npGrid, unsigned int nWidth, unsigned int i, unsigned int j, int bCellState)
{
    unsigned int    nByteNumber, nBitNumber;
    char            c;

    nBitNumber = ((j * nWidth) + i) & 7;
    nByteNumber = ((j * nWidth) + i) >> 3;
    //get current byte state
    c = npGrid[(int)nByteNumber];
    //set new state
    if (bCellState)
        c |= (1 << nBitNumber);
    else
        c &= (~(1 << nBitNumber));
    //insert the byte back into matrix
    npGrid.set_at(nByteNumber,c);
}
