/** =================================================================\n
 *
 * Curve Compression Project - CS 7491 2004 project 1\n
 * @file Point.h
 *
 * Primary Authors: Nguyen Truong (nltruong@cc.gatech.edu)\n
 *                  Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright 2004\n
 * (C) Copyright Georgia Institute of Technology, 2004\n
 *
 * ==================================================================\n
 * Basic Point class definition
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2004-01-24
 * @author Nguyen Truong & Howard Zhou
 *
 *
 * @todo
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#ifndef __CPOINT_H__
#define __CPOINT_H__

#pragma warning(disable : 4786)


#include <iostream>
#include <fstream>

// #include "common.h"

/// @namespace using standard namespace
using namespace std;

// ==================================================================================
/**
 * @class CPoint
 * @brief General Point class definition
 * @ingroup Curve
 *
 * @par requirements
 *
 * @par license
 * This code is absolutely free to use and modify. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 *
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2004-01-24
 * @author Nguyen Truong & Howard Zhou
 *
 * This class is a general point class.
 * interface.
 * 
 * @todo
 * @bug
 * @warning
 */
// ================================================================================


/// Representation of a point in 2D
template <class T>
class CPoint2D
{
private:
	// ------------------------------------------------------
	// private helpers
	// ------------------------------------------------------

public:
	T _x;
	T _y;

	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------
	/// default constructor
	CPoint2D(){};

	/// copy constructor
	CPoint2D(const CPoint2D<T> & p)
	{
		_x = p._x;
		_y = p._y;
	}
		
	CPoint2D(T x, T y)
	{
		_x = x;
		_y = y;
	}
	
	/// destructor
	~CPoint2D(){};

	// ------------------------------------------------------
	// Predicators
	// ------------------------------------------------------
	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------

	// ------------------------------------------------------
	// Overloaded Operators 
	// ------------------------------------------------------
	CPoint2D<T>& operator=  (const  CPoint2D<T>& p);
	CPoint2D<T>  operator-  (void)  const;
	CPoint2D<T>  operator+  (const  CPoint2D<T>& p) const;
	CPoint2D<T>  operator-  (const  CPoint2D<T>& p) const;
	CPoint2D<T>  operator*  (double d) const;
	CPoint2D<T>  operator/  (double d) const;
	CPoint2D<T>& operator+= (const  CPoint2D<T>& p);
	CPoint2D<T>& operator-= (const  CPoint2D<T>& p);
	CPoint2D<T>& operator*= (double d);
	CPoint2D<T>& operator/= (double d);
	bool		 operator== (const  CPoint2D<T>& p) const;
	bool		 operator!= (const  CPoint2D<T>& p) const;
	bool		 operator<  (const  CPoint2D<T>& p) const;	

	// ------------------------------------------------------
	// Input, Output, Test
	// ------------------------------------------------------
	void readFile(FILE* fp);

	// ------------------------------------------------------
	// Other operation
	// ------------------------------------------------------

	// ------------------------------------------------------
	// Friend functions
	// ------------------------------------------------------
	
	/**
	 * operator <<:
	 * Display a 2D point through the ostream
	 * @param os reference to the ostream object
	 * @param p reference to the point
	 * @return ostream with the new stream of curve contents attached
	 */
	friend ostream& operator<< (ostream& os, CPoint2D<T>& p)
	{
		os << "(" << p._x << "," << p._y << ")";
		return os;
	};

};

// function template specialization for double
template <>
void CPoint2D<double>::readFile(FILE* fp)
{
	if(NULL == fp)
	{
		cout << "Error -- NULL file pointer\n";
		exit(-1);
	}

	if (2 != fscanf(fp, "%lf %lf\n", &_x, &_y))
	{
		cout << "Error -- wrong vertex\n";
		exit(-1);
	}
};


// function template specialization for int
template <>
void CPoint2D<int>::readFile(FILE* fp)
{
	if(NULL == fp)
	{
		cout << "Error -- NULL file pointer\n";
		exit(-1);
	}

	if (2 != fscanf(fp, "%d %d\n", &_x, &_y))
	{
		cout << "Error -- wrong vertex\n";
		exit(-1);
	}			
};



// ------------------------------------------------------
// Overloaded Operators 
// ------------------------------------------------------
template <class T>
CPoint2D<T>& CPoint2D<T>::operator=  (const  CPoint2D<T>& p)
{
	_x = p._x;
	_y = p._y;
	return *this;
}

template <class T>
CPoint2D<T>  CPoint2D<T>::operator-  (void) const
{
	return CPoint2D<T>(-_x, -_y); 
}

template <class T>
CPoint2D<T>  CPoint2D<T>::operator+  (const  CPoint2D<T>& p) const
{
	return CPoint2D<T>(_x + p._x, _y + p._y);
}

template <class T>
CPoint2D<T>  CPoint2D<T>::operator-  (const  CPoint2D<T>& p) const
{
	return CPoint2D<T>(_x - p._x, _y - p._y);
}

template <class T>
CPoint2D<T>  CPoint2D<T>::operator*  (double d) const
{
	return CPoint2D<T>(_x*d, _y*d);
}

template <class T>
CPoint2D<T>  CPoint2D<T>::operator/  (double d) const
{
	return CPoint2D<T>(_x/d, _y/d);
}

template <class T>
CPoint2D<T>& CPoint2D<T>::operator+= (const  CPoint2D<T>& p)
{
	_x += p._x;
	_y += p._y;
	return *this;
}

template <class T>
CPoint2D<T>& CPoint2D<T>::operator-= (const  CPoint2D<T>& p)
{
	_x -= p._x;
	_y -= p._y;
	return *this;
}

template <class T>
CPoint2D<T>& CPoint2D<T>::operator*= (double d)
{
	_x *= d;
	_y *= d;
	return *this;
}

template <class T>
CPoint2D<T>& CPoint2D<T>::operator/= (double d)
{
	_x /= d;
	_y /= d;
	return *this;
}

template <class T>
bool CPoint2D<T>::operator== (const  CPoint2D<T>& p) const
{
	return (_x == p._x && _y == p._y);
}

template <class T>
bool CPoint2D<T>::operator!= (const  CPoint2D<T>& p) const
{
	return (_x != p._x || _y != p._y);
}

template <class T>
bool CPoint2D<T>::operator<  (const  CPoint2D<T>& p) const
{
	return (_x < p._x && _y < p._x);
}

typedef CPoint2D<double> CPoint2DReal;
typedef CPoint2D<int> CPoint2DInt;

#endif // __CPOINT_H__