/**********************************************
 * CS7491
 * Howard Zhow & Nguyen Truong
 * First 3 parts of project 1
 **********************************************/

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>

#define DEBUG 1

/*************************/
/* Structs helpful to us */
/*************************/

/* Representation of a real point in 2D */
typedef struct real_point_t {
  double x;
  double y;
} real_point;

/* Representation of an integer point in 2D */
typedef struct int_point_t {
  int x;
  int y;
} int_point;

/* Representation of a vector in 2D */
typedef struct vector_t {
  int x;
  int y;
} vector;

/* Representation of a curve with real points */
typedef struct real_curve_t {
  int vertices; /* number of vertices */
  int edges; /* number of edges */
  int x_bits; /* number of bits to be used for x-coor */
  int y_bits; /* number of bits to be used for y_coor */
  double minX; /* min value of x */
  double maxX; /* max value of x in curve */
  double minY;
  double maxY;
  real_point *curve;
} real_curve;

/* Representation of a curve with integer points */
typedef struct int_curve_t {
  int vertices; /* number of vertices */
  int edges; /* number of edges */
  int x_bits; /* number of bits used for x-coor */
  int y_bits; /* number of bits used for y_coor */
  double minX; /* min value of x */
  double maxX; /* max value of x in curve */
  double minY;
  double maxY;
  int_point *curve;
} int_curve;

/********************/
/* Global variables */
/********************/
real_curve inputCurve;
int_curve quantizedCurve;
vector *residues;


/***********************/
/* Function prototypes */
/***********************/
void readInput(FILE *stream);
void quantization();
double a(int x1, int y1, int x2, int y2, int x3, int y3);
double b(int x1, int y1, int x3, int y3, double a);
double c(int x1, int y1, double a, double b);
void collectResidues(int_curve curve);
void cleanup();

int main(int argc, char *argv[]) {
  FILE *inputStream = NULL;
  int i;

  if (argc < 2) {
    fprintf(stderr, "We need two parameters\n");
    fprintf(stderr, "File name and junk\n");
    return(1);
  }

  /* Open input file, then read and store in inputCurve */

  inputStream = fopen(argv[1], "r");
  if(inputStream != NULL) { /* Such a file exists */
    readInput(inputStream);
    fclose(inputStream);
  }

  /* Convert to integer units */
  quantization();

  /* Collect residues */
  collectResidues(quantizedCurve);

  if(DEBUG) {
    printf("Everything is over now, try displaying\n");
    printf("Min x: %f\n", inputCurve.minX);
    printf("Max x: %f\n", inputCurve.maxX);
    printf("Min y: %f\n", inputCurve.minY);
    printf("Max y: %f\n", inputCurve.maxY);
    for(i=0; i < inputCurve.vertices; i++) {
      printf("x: %f, y: %f\n", inputCurve.curve[i].x, 
	     inputCurve.curve[i].y);
    }

    printf("\n\n\n\nAfter quantization\n\n\n");
    for(i=0; i < quantizedCurve.vertices; i++) {
      printf("x: %d, y: %d\n", quantizedCurve.curve[i].x,
	     quantizedCurve.curve[i].y);
    }

    printf("\n\n\n\nAfter collecting residues\n\n\n");
    for(i=0; i < quantizedCurve.vertices - 3; i++) {
      printf("residues collected: %d %d\n", residues[i].x, residues[i].y);
    }
  }


  cleanup();

  return 0;
}


/* Read in the file, recording the number of vertices, the number
 * of edges (if the two are equal, it's a close curve, otherwise
 * it's an open curve), the number of bits used for x-coor, the
 * number of bits used for y-coor, and the x,y-coordinates for
 * all vertices
 */
void readInput(FILE *stream) {
  char *line = NULL;
  size_t read;
  size_t len = 0;
  int counter = 0;

  /* Initialize inputCurve properties */
  inputCurve.minX = LONG_MAX;
  inputCurve.maxX = LONG_MIN;
  inputCurve.minY = LONG_MAX;
  inputCurve.maxY = LONG_MIN;

  /* Read in the first line, should be the number of vertices */
  if((read=getline(&line, &len, stream))!=-1) {
    inputCurve.vertices = atoi(line);
  }

  /* Read in the second line, should be the number of edges */
  if((read=getline(&line, &len, stream))!=-1) {
    inputCurve.edges = atoi(line);
  }

  /* Read in the third line, should be the number of bits for x_coor */
  if((read=getline(&line, &len, stream))!=-1) {
    inputCurve.x_bits = atoi(line);
  }

  /* Read in the fourth line, should be the number of bits for y_coor */
  if((read=getline(&line, &len, stream))!=-1) {
    inputCurve.y_bits = atoi(line);
  }

  /* Malloc space for inputCurve.curve */
  do {
    inputCurve.curve = malloc(inputCurve.vertices * sizeof(real_point));
  } while(inputCurve.curve == NULL);



  /* Put the respective coordinates into our curve */
  while((read=getdelim(&line, &len, ' ', stream))!=-1) {
    if(DEBUG) {
      printf("Retrieved line of length %zu\n", read);
      printf("%s\n", line);
    }
    inputCurve.curve[counter].x = atof(line);
    if(inputCurve.curve[counter].x < inputCurve.minX)
      inputCurve.minX = inputCurve.curve[counter].x;
    if(inputCurve.curve[counter].x > inputCurve.maxX)
      inputCurve.maxX = inputCurve.curve[counter].x;


    /* get input after the comma: y-coors */
    read=getline(&line, &len, stream);
    if(DEBUG) {
      printf("Retrieved line after comma of length %zu\n", read);
      printf("%s\n", line);
    }
    inputCurve.curve[counter].y = atof(line);
    if(inputCurve.curve[counter].y < inputCurve.minY)
      inputCurve.minY = inputCurve.curve[counter].y;
    if(inputCurve.curve[counter].y > inputCurve.maxY)
      inputCurve.maxY = inputCurve.curve[counter].y;

    counter++;
  }

  /*fprintf(stderr, "Error happened here\n");*/

  if(line)
    free(line);

  return;
}


/* Store float coordinates of our curve in integers, using B-bits */
void quantization() {
  int xLength, yLength;
  double boxLengthX, boxLengthY;
  int counter;

  quantizedCurve.vertices = inputCurve.vertices;
  quantizedCurve.edges = inputCurve.edges;
  quantizedCurve.x_bits = inputCurve.x_bits;
  quantizedCurve.y_bits = inputCurve.y_bits;
  quantizedCurve.minX = inputCurve.minX;
  quantizedCurve.maxX = inputCurve.maxX;
  quantizedCurve.minY = inputCurve.minY;
  quantizedCurve.maxY = inputCurve.maxY;

  /* Mallocking space for our representation of a curve with
   * integer points
   */
  do {
    quantizedCurve.curve=malloc(quantizedCurve.vertices *
				sizeof(int_point));
  } while(quantizedCurve.curve == NULL);


  /* Calculating 2^Bx and 2^By */
  xLength = pow(2, (double)quantizedCurve.x_bits);
  yLength = pow(2, (double)quantizedCurve.y_bits);

  boxLengthX = inputCurve.maxX - inputCurve.minX;
  boxLengthY = inputCurve.maxY - inputCurve.minY;

  /* Now represent the curve in integer coordinates */
  for(counter=0; counter < quantizedCurve.vertices; counter++) {
    quantizedCurve.curve[counter].x 
      = (int)floor(((inputCurve.curve[counter].x - inputCurve.minX)/
		    boxLengthX) * xLength);

    quantizedCurve.curve[counter].y
      = (int)floor(((inputCurve.curve[counter].y - inputCurve.minY)/
		    boxLengthY) * yLength);
  }
  return;
}


/* The following 3 functions calculate a, b, and c in the quadratic
 * formula ax^2 + bx + c = y
 */
double a(int x1, int y1, int x2, int y2, int x3, int y3) {
  double y2minusy1 = y2 - y1;
  double x2minusx1 = x2 - x1;
  double x3minusx1 = x3 - x1;
  double y3minusy1 = y3 - y1;
  double x2squareMinusx1square = x2 * x2 - x1 * x1;
  double x3plusx1 = x3 + x1;

  double toreturn;

  toreturn = ((y2minusy1 - (x2minusx1/x3minusx1)*(y3minusy1))/
	      (x2squareMinusx1square - (x2minusx1*x3plusx1)));

  if(isinf(toreturn) || isnan(toreturn))
    return 1;
  else
    return toreturn;
}


double b(int x1, int y1, int x3, int y3, double a) {
  double y3minusy1 = y3 - y1;
  double x3squareMinusx1square = x3 * x3 - x1 * x1;
  double x3minusx1 = x3 - x1;
  double toreturn;

  toreturn = ((y3minusy1 - a*(x3squareMinusx1square))/x3minusx1);

  if(isinf(toreturn) || isnan(toreturn))
    return 1;
  else
    return toreturn;
}


double c(int x1, int y1, double a, double b) {
  return (y1 - b*x1 - a*x1*x1);
}


/* Collect the residues.  Assume that you have the first three
 * vertices of a curve, you can reconstruct the fourth point
 * by using the quadratic predictor, then add the residues
 * to it.  You can recursively get next points from 3 previous
 * points.  This is how we should decode our file.
 */
void collectResidues(int_curve curve) {
  int counter;

  /* malloc space for out residues place holder first */
  do {
    residues = malloc((quantizedCurve.vertices-3)*sizeof(vector));
  } while(residues == NULL);

  /* Loop through all of our vertices, starting from the 3-rd
   * one, and estimates the corresponding residue, using the
   * quadratic predictor
   */
  for(counter=3; counter<quantizedCurve.vertices; counter++) {
    double aVal = a(quantizedCurve.curve[counter-3].x,
		    quantizedCurve.curve[counter-3].y,
		    quantizedCurve.curve[counter-2].x,
		    quantizedCurve.curve[counter-2].y,
		    quantizedCurve.curve[counter-1].x,
		    quantizedCurve.curve[counter-1].y);

    double bVal = b(quantizedCurve.curve[counter-3].x,
		    quantizedCurve.curve[counter-3].y,
		    quantizedCurve.curve[counter-1].x,
		    quantizedCurve.curve[counter-1].y,
		    aVal);

    double cVal = c(quantizedCurve.curve[counter-3].x,
		    quantizedCurve.curve[counter-3].y,
		    aVal, bVal);

    /* The above three double values we just calculated gave us
     * the coefficients for our estimating quadratic formula,
     * now we can use them to estimate the value of the 3rd
     * term
     */
    double yEstimate = aVal*pow(quantizedCurve.curve[counter].x, 2) +
      bVal*quantizedCurve.curve[counter].x + cVal;

    printf("aVal is %f, bVal is %f, cVal is %f\n", aVal, bVal, cVal);

    residues[counter-3].y = (int)yEstimate - 
      quantizedCurve.curve[counter].y;

    residues[counter-3].x = quantizedCurve.curve[counter].x -
      quantizedCurve.curve[counter-1].x;
  }
}


/* free all malloc'ed data, if any was allocated */
void cleanup() {
  if(inputCurve.curve)
    free(inputCurve.curve);

  if(quantizedCurve.curve)
    free(quantizedCurve.curve);

  if(residues)
    free(residues);
  return;
}
