/** =================================================================\n
 *
 * MathLib\n
 * @file Vector.h
 *
 * Primary Author: Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright Howard Zhou 2003\n
 * (C) Copyright Georgia Institute of Technology, 2003\n
 *
 * ==================================================================\n
 * Basic Vector lab
 *
 * @version 1.1\n
 *     Major organization change, moved most of function defintion to\n
 *     the .cpp file to make class declaration more clear.
 * @date 2003-05-14
 * @author howardz
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 *
 * @todo Implement det() and invert() 
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */
#include "Vector.h"


#include <stdio.h>
#include <iomanip>

//#include <math.h>
//#include <assert.h>
//#include <string>

/// @namespace using standard namespace
using namespace std;

// ======================================================
// Actual function definition for CVector
// ======================================================

// ------------------------------------------------------
// Output or Test
// ------------------------------------------------------

/**
 * Member display function\n
 * Display the current vector
 */
void CVector::display() const
{
	assert(isValid());
	printf("\n[");
	for(int i=0; i<_len; i++)
		printf("%f ", _v[i]);
	printf("]\n");
}


/**
 * General test function for CMatrix class
 * put anything you want to test here.
 */
void CVector::test() const
{
	cout << "\n\nCVector Test\n=========================\n";

	CVector v1(4, 1);
	v1.display();
	CVector v2(4, 2);
	v2.display();

	v1 = v2;
	v1.display();
	v2 = 3;
	v2[1] = -2;
	v2.display();
	v1 = -v2;
	v1.display();
	v2.display();
	printf("length %d, magnitude %f\n", v2.length(), v2.magnitude());
	v2.normalize();
	v2.display();

	CVector v3(4, 5);
	v3[1] = 0;
	v3[2] = 0;
	v3[3] = 12;
	v3.display();
	printf("length %d, magnitude %f\n", v3.length(), v3.magnitude());
	v3.normalize();
	printf("length %d, magnitude %f\n", v3.length(), v3.magnitude());
	v3.display();
	CVector v31(v3.normalized());
	printf("length %d, magnitude %f\n", v31.length(), v31.magnitude());
	cout << "\nv3  : " << v3;
	cout << "\nv31 : " << v31; 


	v3[0] = 0;
	v3[3] = 0;
	v3[2] = 1;
	printf("length %d, magnitude %f\n", v3.length(), v3.magnitude());
	v3.normalize();
	printf("length %d, magnitude %f\n", v3.length(), v3.magnitude());
	v3.display();		
	
	v3[0] = 5;
	v3[1] = 0;
	v3[2] = 6;
	v3[3] = 12;

	cout << v3 << std::endl;

	printf("+, +=, -, -=, *, *=, /, /=\n");
	v1= CVector(4, 1);
	cout << v1 << std::endl;
	v2= CVector(4);
	cout << v2 << std::endl;

	CVector v4 = v1+v2;
	cout << v4 << std::endl;
	v4 += v1;
	cout << v4 << std::endl;
	v4 = v1 - v2;
	cout << v4 << std::endl;
	v4 -= v1;
	cout << v4 << std::endl;

	cout << v3 << std::endl;
	printf("v3 * v3 = %f\n", v3 * v3);
	v4 = CVector(4, 1);
	cout << v4 << std::endl;
	v4 *= 2;
	cout << v4 << std::endl;
	v4 = v4 - v1;
	cout << v4 << std::endl;
	v4 -= v3;
	cout << v4 << std::endl;
	cout << v2 << std::endl;

	if(v2 == v4)
		printf("equal");
	else
		printf("not equal");
}

// ------------------------------------------------------
// Friend functions
// ------------------------------------------------------

/**
 * Provides special formatting when a matrix or vector (respectively) is part of the output stream.
 * @param os reference to the output stream	
 * @param refvector reference to a vector
 * @return reference to the output stream
 */
ostream& operator<< (ostream& os, CVector& refvector)
{
	assert(refvector.isValid());
	os << " [ ";
	for(int i=0; i<refvector._len; i++)
	{
		using namespace std;
		os << setprecision(3);
		os << setw(8);
		os << setiosflags (ios_base::fixed | ios_base::right);
		os << refvector._v[i] << " ";
	}
	os << "] ";
	return os;
}


/**
 * Function:	lerp\n
 * Purpose:		Linear Interpolation Between two Vectors\n
 * Adjust the DELTA constant to play with the LERP vs. SLERP level\n
 * @param v1 vector
 * @param v2 vector
 * @param blend blend factor
 * @return result vector which is a linear interpolation between two input vectors
 */
CVector lerp(CVector v1, CVector v2, float blend)
{
	// make sure that blend is in [0, 1]
	assert(blend >= 0 && blend <= 1);
	CVector result = v1*(1-blend) + v2 *(blend);
	return CVector(result);
}


// ======================================================
// Actual function definition for CVector3
// ======================================================

// ------------------------------------------------------
// Output or Test
// ------------------------------------------------------

/**
 * General test function for CMatrix class
 * put anything you want to test here.
 */
void CVector3::test() const
{
	cout << "\n\nCVector3 Test\n=========================\n";
	CVector3 v1(1, 2, 3);
	CVector3 v2(4, 5, 6);
	CVector3 v3 = v1.cross(v2);
	cout << v3;

	v3 = v2.cross(v1);
	cout << v3;
}

// ======================================================
// Actual function definition for CVector4
// ======================================================

// ------------------------------------------------------
// Output or Test
// ------------------------------------------------------

/**
 * General test function for CMatrix class
 * put anything you want to test here.
 */
void CVector4::test() const
{
	cout << "\n\nCVector4 Test\n=========================\n";
}

//@}