/** =================================================================\n
 *
 * MathLib\n
 * @file Vector.h
 *
 * Primary Author: Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright Howard Zhou 2003\n
 * (C) Copyright Georgia Institute of Technology, 2003\n
 *
 * ==================================================================\n
 * Basic Vector lab
 *
 * @version 1.1\n
 *     Major organization change, moved most of function defintion to\n
 *     the .cpp file to make class declaration more clear.
 * @date 2003-05-14
 * @author howardz
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 *
 * @todo Implement det() and invert() 
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#ifndef ___VECTOR__H___
#define ___VECTOR__H___

#include <assert.h>
#include <iostream>
#include <math.h>


//#include <stdio.h>
//#include <string>
//#include <iomanip>

/// @namespace using standard namespace
using namespace std;


// ==================================================================================
/**
 * @class CVector Vector.h Vector.h
 * @brief General Vector class
 * 
 * @ingroup Lib
 *
 * @par requirements
 * MFC\n
 *
 * @par license
 * This code is absolutely free to use and modify. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 * 
 * @version 1.1\n
 *     Major organization change, moved most of function defintion to the .cpp\n
 *     file to make class declaration more clear.
 * @date 2003-05-14
 * @author howardz
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 *
 * @todo none
 * @bug none
 * @warning none
 */
// ================================================================================
class CVector
{
protected:
	int _len;         ///< the length of the vector
	double *_v;       ///< pointer to the memory holding the data

public:

public:
	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------
	CVector     (void);
	CVector     (const CVector &vec);
	CVector     (int length, double elValue);
	CVector     (int length, double *storage = 0);
	void reSize (int length, double *storage = 0);

	~CVector (void);	// Destructor

	// ------------------------------------------------------
	// Predicators
	// ------------------------------------------------------
	bool isValid        (void) const;
	bool isCompatible   (const CVector& refvector) const;

	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------
	int      length     (void) const;
	double	 magnitude  (void) const;
    CVector& normalize  (void);
    CVector  normalized (void) const; 
	
	// ------------------------------------------------------
	// Overloaded Operators 
	// ------------------------------------------------------
	double&  operator[] (int i);
	double   operator[] (int i) const;
	CVector& operator=  (double value); 
	CVector& operator=  (const  CVector& refvector);
	CVector  operator-  (void)  const;
	CVector  operator+  (const  CVector& refvector) const;
	CVector  operator-  (const  CVector& refvector) const;
	CVector  operator*  (double dbl) const;
	double   operator*  (const  CVector& refvector) const;
	CVector  operator/  (double dbl) const;
	CVector& operator+= (const  CVector& refvector);
	CVector& operator-= (const  CVector& refvector);
	CVector& operator*= (double dbl);
	double   operator*= (const  CVector& refvector);
	CVector& operator/= (double dbl);
	bool     operator== (const  CVector& refvector) const;
	bool     operator!= (const  CVector& refvector) const;
	bool     operator<  (const  CVector& refvector) const;


	// ------------------------------------------------------
	// Output or Test
	// ------------------------------------------------------
	void     display() const;
	void	    test() const;

	// ------------------------------------------------------
	// Friend functions
	// ------------------------------------------------------
	friend std::ostream& operator<< (std::ostream& os, CVector& refvector);
	friend CVector lerp(CVector v1, CVector v2, float blend);
};


// ==================================================================================
/**
 * @class CVector3 Vector.h Vector.h
 * @brief 3-Dimensional Vector class
 *
 * @ingroup Lib
 *
 * @par requirements
 * MFC\n
 *
 * @par license
 * This code is absolutely free to use and modify. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 * 
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 *
 * @todo none
 * @bug none
 * @warning none
 */
// ================================================================================
class CVector3 : public CVector
{
public:
	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------
	CVector3 (void);
	CVector3 (const CVector3 &refvector);
	CVector3 (const CVector  &refvector);
	CVector3 (double x, double y, double z);

	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------
	void set (double x, double y, double z);
	CVector3 cross(const CVector3 &refvector) const;

	// ------------------------------------------------------
	// Output or Test
	// ------------------------------------------------------
	void test(void) const;
};

typedef CVector3 CPoint3D;

// ==================================================================================
/**
 * @class CVector4 Vector.h Vector.h
 * @brief 4-Dimensional Vector class
 *
 * @ingroup Lib
 *
 * @par requirements
 *
 * @par license
 * This code is absolutely free to use and modify. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 * 
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 *
 * @todo none
 * @bug none
 * @warning none
 */
// ================================================================================
class CVector4: public CVector
{
public:
	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------
	CVector4 (void);
	CVector4 (const CVector4 &refvector);
	CVector4 (const CVector  &refvector);
	CVector4 (double x, double y, double z, double w);

	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------
	void set (double x, double y, double z, double w);

	// ------------------------------------------------------
	// Output or Test
	// ------------------------------------------------------
	void test() const;
};



/////////////////////////////////////////////////////////
/////////////////// implementation //////////////////////
/////////////////////////////////////////////////////////


// ======================================================
// Actual function definition for CVector
// ======================================================

// ------------------------------------------------------
// constructors and destructors
// ------------------------------------------------------

/**
 * Default Constructor\n
 * Creates a vector instance, but does not initialize it.
 */
inline CVector::CVector (void)
{
	_len = -1;
	_v = 0;
}

/**
 * Copy Constructor\n
 * Creates a vector instance with the number of elements and 
 * element values of vector vec.
 * @param &vec reference to a vector
 */
inline CVector::CVector (const CVector &vec)
{
	_len = vec._len;

	_v = new double[_len];
	memcpy(_v, vec._v, _len * sizeof(double));
}
	
/**
 * Constructor\n
 * Creates a vector instance with length l and all of its elements
 * elValue
 * @param length the lenght of the vector
 * @param elValue the value to assign to the vector
 */
inline CVector::CVector (int length, double elValue)
{
	assert(length > 0);
	_len = length;

	_v = new double[length];

	if(0 == elValue)
		//use memset when ever it is possible
		memset(_v, 0, length*sizeof(double));
	else
		for(int i=0; i< _len; i++)
			_v[i] = elValue;
}

/**
 * Constructor\n
 * Creates a dynamically sized double-precision vector with length elements.
 * If storage is zero, new memory is allocated. If storage is not zero, use 
 * the memory pointed to by storage.
 * @param length the dimention of the vector
 * @param *storage point to the memory
 */
inline CVector::CVector(int length, double *storage)
{
	assert(length > 0);

	_len = length;

	if(0 == storage)
	{
		_v = new double[length];
		memset( _v, 0, _len*sizeof(double) ) ; 
	}
	else
		_v = storage;
}

/**
 * If storage is not zero, reorganizes the memory storage 
 * points to to have length number of elements. If storage 
 * is zero, the appropriate amount of memory is allocated.
 * @param length the dimention of the vector
 * @param *storage pointer to a existing memory
 */
inline void CVector::reSize (int length, double *storage)
{
	if(_v)
		delete[] _v ; 

	_len = length;

	if(0 == storage)
	{
	    _v = new double[length] ;
        memset( _v, 0, length * sizeof(double) ) ; 
	}
	else
	{
		_v = storage;
	}
}

/**
 * Destructor
 */
inline CVector::~CVector()
{
	if(_v)
		delete[] _v;
}


// ------------------------------------------------------
// Predicators
// ------------------------------------------------------
/**
 * A vector is valid if it is initialized and it's length is
 * positive
 * @return true if the vector is valid
 */
inline bool CVector::isValid() const
{
	return (_len > 0 && _v != 0);
}

/**
 * Check if the refvector is compatible with the current vector
 * @param refvector reference to a vector
 * @return true if the vector is valid and of the same dimention
 * as the current vector, false otherwise
 */
inline bool CVector::isCompatible(const CVector& refvector) const
{
	return (isValid() && refvector.isValid() && refvector._len == _len);
}

// ------------------------------------------------------
// Accessors and Mutators
// ------------------------------------------------------

/**
 * Length: dimention of the vector
 * @return int: length (dimention) of vector (number of elements).
 */
inline int CVector::length() const
{
	return _len;
}

/**
 * Magnitude of the vector
 * @return double: the magnitude of the vector
 */
inline double CVector::magnitude() const
{
	assert(isValid());
	double sum = 0;

	for(int i = 0; i<_len; i++)
		sum += _v[i]*_v[i];
	return sqrt(sum);
}

/**
 * Normalize the vector \n
 * v = v/v.magnitude()
 * @return a reference to the vector
 */
inline CVector& CVector::normalize()
{
	assert(isValid());
	double mag = magnitude();

	for(int i= 0; i<_len; i++)
		_v[i] /= mag;
	return (*this);
}

/**
 * Returns a vector that is the normalized version of the current vector.
 * @return a vector that is the normalized version of the current vector.
 */
inline CVector CVector::normalized() const
{
	CVector vec(*this);
	return CVector(vec.normalize());
}

// ------------------------------------------------------
// Overloaded Operators 
// ------------------------------------------------------
 
/**
 * operator [] \n
 * @param i index
 * @return the refernce to the current vecotr element
 */
inline double& CVector::operator[] (int i)
{
	assert(isValid());
	assert(i >= 0 && i< _len);
	
	return _v[i];
}

/**
 * operator [] (Read only) \n
 * Returns vector element value specified by index i (Read Only).
 * @param i index
 * @return the value of the current vector element
 */
inline double CVector::operator[] (int i) const
{
	assert(isValid());
	assert(i >= 0 && i< _len);
	
	return _v[i];
}

/**
 * operator = \n
 * Assigns value to all elements of this vector.
 * @param value value to be assigned
 * @return reference to the current vector 
 */
inline CVector& CVector::operator= (double value) 
{
	assert(isValid());
	for(int i = 0; i<_len; i++)
		_v[i] = value;

	return (*this);
}

/**
 * operator = (const CVector& vec) \n
 * Copies the values of another vector to the 
 * corresponding elements of this vector.
 * @param refvector 
 * @return reference to the current vector 
 */
inline CVector& CVector::operator= (const CVector& refvector) 
{
	assert(isCompatible(refvector));	
	_len = refvector._len;

	memcpy(_v, refvector._v, _len * sizeof(double));

	return (*this);
}	

/**
 * operator - [unary] \n
 * @return a vector in which the elements of 
 * the current vector have been multiplied by -1
 */
inline CVector CVector::operator- (void) const 
{
	assert(isValid());
	CVector vec(_len);

	for(int i = 0; i<_len; i++)
		vec._v[i] = -_v[i];

	return CVector(vec);
}

/**
 * operator + \n
 * @param refvector reference to a vector 
 * @return a vector that is the sum of the current vector and refvector.
 */
inline CVector CVector::operator+ (const CVector& refvector) const 
{
	assert(isCompatible(refvector));
	CVector vec(_len);

	for(int i = 0; i<_len; i++)
		vec._v[i] = _v[i] + refvector._v[i];

	return CVector(vec);
}

/**
 * operator - \n
 * @param refvector reference to a vector  
 * @return a vector that is the difference of the current vector and refvector.
 */
inline CVector CVector::operator- (const CVector& refvector) const 
{
	assert(isCompatible(refvector));
	CVector vec(_len);

	for(int i = 0; i<_len; i++)
		vec._v[i] = _v[i] - refvector._v[i];

	return CVector(vec);
}

/**
 * operator * (dot product with refvecotr) \n
 * @param &refvector reference to a vector  
 * @return the dot product of refvector and current vector.
 */
inline double CVector::operator* (const CVector &refvector) const
{
	assert(isCompatible(refvector));
	double dot_prod = 0;

	for(int i = 0; i<_len; i++)
		dot_prod += _v[i] * refvector._v[i];

	return dot_prod;
}

/**
 * operator * (multiplication by double) \n
 * @param dbl double as a mulitplyer
 * @return a vector that is the current vector multiplied the dbl value.
 */
inline CVector CVector::operator* (double dbl) const
{
	assert(isValid());
	CVector vec(_len);

	for(int i = 0; i<_len; i++)
		vec._v[i] = _v[i] * dbl;

	return CVector(vec);
	
}

/**
 * operator / (division by double)\n
 * @param dbl double as a divider
 * @return a vector that is the current vector divided the dbl value. 
 */
inline CVector CVector::operator/ (double dbl) const 
{
	assert(isValid());
	CVector vec(_len);

	for(int i = 0; i<_len; i++)
		vec._v[i] = _v[i] / dbl;

	return CVector(vec);	
}

/**
 * operator += \n
 * Adds the elements of refvector to the corresponding elements of the current vector.
 * @param refvector reference to a vector  
 * @return reference to the current vector
 */
inline CVector& CVector::operator+= (const CVector& refvector) 
{
	assert(isCompatible(refvector));
	for(int i=0; i < _len; i++)
		_v[i] += refvector._v[i];
	return *this;
}

/**
 * operator -= \n
 * Subtracts the elements of refvector from the corresponding elements of the current vector.
 * @param refvector reference to a vector  
 * @return reference to the vector
 */
inline CVector& CVector::operator-= (const CVector& refvector) 
{
	assert(isCompatible(refvector));
	for(int i=0; i < _len; i++)
		_v[i] -= refvector._v[i];
	return *this;
}

/**
 * operator *=\n
 * Multiplies the elements of the current vector by dbl.
 * @param dbl double as a multiplyer
 * @return reference to the vector
 */
inline CVector& CVector::operator*= (double dbl) 
{
	assert(isValid());
	for(int i=0; i < _len; i++)
		_v[i] *= dbl;
	return *this;
}

/**
 * operator /= \n
 * Divides the elements of the current vector by dbl.
 * @param dbl double as a divider
 * @return reference to the vector
 */
inline CVector& CVector::operator/= (double dbl) 
{
	assert(isValid());
	for(int i=0; i < _len; i++)
		_v[i] /= dbl;
	return *this;
}

/**
 * operator ==\n
 * @param refvector reference to a vector  
 * @return true if all elements of refvector are equal 
 * to the corresponding elements of the current vector.
 */
inline bool CVector::operator== (const CVector& refvector) const 
{
	assert(isCompatible(refvector));
	for(int i=0; i < _len; i++)
		if(_v[i] != refvector._v[i])
			return false;
	return true;
}

/**
 * operator !=\n
 * @param refvector reference to a vector  
 * @return true if any element of refvector is not equal 
 * to the corresponding element of the current vector.
 */
inline bool CVector::operator!= (const CVector& refvector) const 
{
	assert(isCompatible(refvector));
	for(int i=0; i < _len; i++)
		if(_v[i] != refvector._v[i])
			return true;
	return false;
}


/**
 * operator <\n
 * This operator is defined so this class can be used in 
 * STL containers.
 * @param refvector reference to a vector  
 * @return true if the magnitude of the refvector is smaller
 * than the current one.
 */
inline bool CVector::operator< (const CVector& refvector) const 
{
	assert(isCompatible(refvector));
	return (magnitude() <  refvector.magnitude());
}


// ======================================================
// Actual function definition for CVector3
// ======================================================

/**
 * Default Constructor \n
 */
inline CVector3::CVector3():CVector(3){}

/**
 * Constructor \n
 * @param x 
 * @param y 
 * @param z 
 */
inline CVector3::CVector3(double x, double y, double z):CVector(3)
{
	_v[0] = x;
	_v[1] = y;
	_v[2] = z;
}

/**
 * Copy Constructor \n
 * @param &refvector reference to a vector 
 */
inline CVector3::CVector3 (const CVector3 &refvector):CVector(refvector){}
		
/**
 * constructor (type casting)\n
 * @param &refvector reference to a vector
 */
inline CVector3::CVector3 (const CVector &refvector)
{
	assert(3 == refvector.length());

	_len = 3;
	_v = new double[3];

	_v[0] = refvector[0];
	_v[1] = refvector[1];
	_v[2] = refvector[2];
}

/**
 * Sets the value of the current vector
 * @param x 
 * @param y 
 * @param z 
 */
inline void CVector3::set(double x, double y, double z)
{
	_v[0] = x; _v[1] = y; _v[2] = z;
}

/**
 * Cross product \n
 * @param &refvector reference to a vector
 * @return the cross product between the two vectors
 */
inline CVector3 CVector3::cross(const CVector3 &refvector) const
{

	CVector3 vec;
	vec._v[0] = _v[1]*refvector._v[2] - _v[2]*refvector._v[1];
	vec._v[1] = _v[2]*refvector._v[0] - _v[0]*refvector._v[2];
	vec._v[2] = _v[0]*refvector._v[1] - _v[1]*refvector._v[0];
	
	return CVector3(vec);
}

// ======================================================
// Actual function definition for CVector3
// ======================================================

/**
 * Default Constructor \n
 */
inline CVector4::CVector4():CVector(4){};

/**
 * Constructor \n
 * @param x 
 * @param y 
 * @param z 
 * @param w 
 */
inline CVector4:: CVector4(double x, double y, double z, double w):CVector(4)
{
	_v[0] = x;
	_v[1] = y;
	_v[2] = z;
	_v[3] = w;
}

/**
 * Copy Constructor \n
 * @param &refvector reference to a vector
 */
inline CVector4:: CVector4 (const CVector4 &refvector):CVector(refvector){}

/**
 * constructor (type casting CVector into CVector4)\n
 * @param &refvector reference to a vector
 */
inline CVector4:: CVector4 (const CVector &refvector)
{
	assert(4 == refvector.length());

	_len = 4;
	_v = new double[4];

	_v[0] = refvector[0];
	_v[1] = refvector[1];
	_v[2] = refvector[2];
	_v[3] = refvector[3];
}

/**
 * Sets the value of the current vector
 * @param x 
 * @param y 
 * @param z 
 * @param w 
 */
inline void CVector4::set(double x, double y, double z, double w)
{
	_v[0] = x; _v[1] = y; _v[2] = z; _v[3] = w;
}


#endif
//@}