/** =================================================================\n
 *
 * TriangleMesh Compression Project - CS 7491 2004 project 1\n
 * @file TriangleMesh.cpp
 *
 * Primary Authors: Brandon Beck
 *					Nguyen Truong (nltruong@cc.gatech.edu)\n
 *                  Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright 2004\n
 * (C) Copyright Georgia Institute of Technology, 2004\n
 *
 * ==================================================================\n
 * Basic TriangleMesh class definition
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2004-02-14
 * @author Brandon Beck, Nguyen Truong, Howard Zhou
 *
 *
 * @todo
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#include <assert.h>
#include <algorithm>
#include <fstream>

#include "TriangleMesh.h"

/// @namespace using standard namespace
using namespace std;


#define NEXT(c) (((c)%3 == 2)?((c)-2):((c)+1))
#define PREVIOUS(c) (((c)%3 == 0)?((c)+2):((c)-1))


class CTriplet
{
public:
	long _t1;
	long _t2;
	long _t3;

	CTriplet(long t1, long t2, long t3)
	{
		_t1 = t1;
		_t2 = t2;
		_t3 = t3;
	}

	// operator < defined for sorting
	bool operator < (const CTriplet & ref)
	{
		if(_t1 < ref._t1)
		{
			return true;
		}
		else if(_t1 == ref._t1)
		{
			if(_t2 < ref._t2)
			{
				return true;
			}
			else if(_t2 == ref._t2)
			{
				if(_t3 < ref._t3)
				{
					return true;
				}
			}
		}
		return false;
	}

	friend ostream & operator << (ostream & os, const CTriplet & ref)
	{
		os << " (" << ref._t1 << ", " << ref._t2 << ", " << ref._t3 << ") ";
		return os;
	}

};



/// local helper function declaration
/*
bool orientationTest(CPoint3D c, CPoint3D p, CPoint3D q, CPoint3D r);
double twoLinEqs(double a, double b, double c, double d, double e, double f);
CPoint3D circumscribingCenter(CPoint3D p, CPoint3D q, CPoint3D r);
CPoint3D sphereCenter(CPoint3D p, CPoint3D q, CPoint3D r, double radius, bool positiveSign);
bool sphereCenters(CPoint3D p, CPoint3D q, CPoint3D r, double radius, CPoint3D& pos, CPoint3D& neg);
*/

template <class T>
void displayVector(vector<T> list)
{
	vector<T>::iterator curr;

	cout << "\n-------------------\n";
	for(curr = list.begin();
		curr != list.end();
		curr ++)
	{
		cout << (*curr) << " -- ";
	}
	cout << "\n-------------------\n";
}

// ------------------------------------------------------
// private helpers
// ------------------------------------------------------


// ------------------------------------------------------
// constructors and destructors
// ------------------------------------------------------

// ------------------------------------------------------
// Predicators
// ------------------------------------------------------

// ------------------------------------------------------
// Accessors and Mutators
// ------------------------------------------------------

// ------------------------------------------------------
// Overloaded Operators 
// ------------------------------------------------------

// ------------------------------------------------------
// Input, Output, and Test
// ------------------------------------------------------
void CTriangleMesh::readFile(const char* filename)
{
	// clean up the original triangle
	_geometryTable.clear();
	_triangleTable.clear();
	_oppositeTable.clear();
	_shellTable.clear();

	FILE * inFile;
	char line[255];
	long i, nT, nV, vi, oi;
	double vx, vy, vz;

	// open the file
	if(NULL == (inFile = fopen(filename, "rb")))
	{
		cout << "Error -- unable to open file <" << filename << ">\n";
		exit(-1);
	}

	// get the triangle count
	if(NULL == fgets(line, 255, inFile) || (nT = atol(line)) == 0)
	{
		cout << "Error -- wrong file format\n";
		exit(-1);
	}

	// corner table and opposite table
	for(i = 0; i < nT*3; i++)
	{
		fscanf(inFile, "%ld %ld\n", &vi, &oi);
		_triangleTable.push_back(vi);
		_oppositeTable.push_back(oi);
	}

	// get the vertex count
	if(NULL == fgets(line, 255, inFile) || (nV = atol(line)) == 0)
	{
		cout << "Error -- wrong file format\n";
		exit(-1);
	}

	// get vertex coordinates
	for(i = 0; i < nV; i++)
	{
		fscanf(inFile, "%lf %lf %lf\n", &vx, &vy, &vz);
		_geometryTable.push_back(CPoint3D(vx, vy, vz));
	}

	fclose(inFile);
}

void CTriangleMesh::writeFile	(const char* filename) const
{
	long i, nT, nV;
	FILE * outFile;
	CPoint3D p;

	if(NULL == (outFile = fopen(filename, "w")))
	{
		cout << "Error -- unable to open file <" << filename << ">\n";
		exit(-1);
	}

	nT = long(getTriangleCount());
	assert(nT == (long)_oppositeTable.size()/3);

	nV = long(_geometryTable.size());

	fprintf(outFile, "%ld\n", nT);

	for(i = 0; i < nT*3; i++)
	{
		fprintf(outFile, "%ld %ld\n", _triangleTable[i], _oppositeTable[i]);
	}

	fprintf(outFile, "%ld\n", nV);

	for(i = 0; i < nV; i++)
	{
		p = _geometryTable[i];
		fprintf(outFile, "%f %f %f\n", float(p[0]) ,float(p[1]), float(p[2]));
	}

	fclose(outFile);
}



void CTriangleMesh::test()
{
	cout << "# vertices  : " << (long)_geometryTable.size() << endl;
	cout << "# triangles : " << (long)_triangleTable.size() << endl;
	displayVector(_triangleTable);
	displayVector(_geometryTable);

	/*
	cout << (int)orientationTest(CPoint3D(0,0,0), CPoint3D(1,0,0), CPoint3D(1,1,1), CPoint3D(0,1,0)) << endl;
	cout << (int)orientationTest(CPoint3D(2,2,0), CPoint3D(1,0,0), CPoint3D(1,1,1), CPoint3D(0,1,0)) << endl;

	cout << circumscribingCenter(CPoint3D(-2,-1,0), CPoint3D(-1,0,0), CPoint3D(0,-1,0)) << endl;
	cout << sphereCenter(CPoint3D(-2,-1,0), CPoint3D(-1,0,0), CPoint3D(0,-1,0), 1, 0);
	cout << sphereCenter(CPoint3D(-2,-1,0), CPoint3D(-1,0,0), CPoint3D(0,-1,0), 2, 0);
	*/
}

void CTriangleMesh::displayOppositeTable(void)
{

	vector<long>::iterator curr;
	
	ofstream flog("log.txt", ios::app);

	flog << endl;
	for(curr = _oppositeTable.begin(); curr != _oppositeTable.end(); curr++)
	{
		flog << " " << *curr;
	}
	flog << endl;
	flog.close();
	
	/*
	cout << "opposite table size : " << (long)_oppositeTable.size() << endl;

	cout << endl;
	for(curr = _oppositeTable.begin(); curr != _oppositeTable.end(); curr++)
	{
		cout << " " << *curr;
	}
	cout << endl;
	*/
}



void CTriangleMesh::displayCornerTable(void)
{
	long i;
	ofstream flog("corner.txt", ios::out);

	for(i = 0; i < (long)_triangleTable.size(); i++)
	{
		flog << i << "\t" << _triangleTable[i] << "\t" <<  _oppositeTable[i] << endl;
	}
	flog.close();
	
}
// ------------------------------------------------------
// Other operation
// ------------------------------------------------------


void CTriangleMesh::addVertices(vector<CPoint3D> points)
{
	_geometryTable.insert(_geometryTable.end(), points.begin(), points.end());
}


/**
 * constructOppositeTable
 * O(nlogn)
 * 
 **/
bool CTriangleMesh::buildOppositeTable(void)
{
	if (0 == _triangleTable.size())
		return false;

	vector<CTriplet> triplets;
	long i, j, c,  cp, cn, cpv, cnv, c1, c2;
	bool bNoMatch = false;

	// initialize the opposite table using triangle table since
	// they have the same length
	_oppositeTable = _triangleTable;

	// make all the triplets
	for(c = 0; c < (long)_triangleTable.size(); c++)
	{
		_oppositeTable[c] = -1;

		cn = NEXT(c);
		cp = PREVIOUS(c);
		
		cpv = _triangleTable[cp];
		cnv = _triangleTable[cn];
		
		triplets.push_back(CTriplet(min(cnv, cpv), max(cnv, cpv), c));
	}

	//cout << triplets.size() << endl;

	// sort the triplet
	sort(triplets.begin(), triplets.end());

	// a bin is an abstraction of a sequence of triplets that contains
	// the same first two elements

	// construct the opposite table
	for(i = 0; i < (long)triplets.size()-1 ;i++)
	{
		// for each pair of triplets
		c1 = triplets[i]._t3;

		if (c1 == 222)
			c1 = 222;


		if(_oppositeTable[c1] != -1)
			continue;


		// cout << "\n" << triplets[i] << "---";

		for (j = i+1; j < (long)triplets.size(); j++)
		{
			c2 = triplets[j]._t3;

			// cout << triplets[j];
			long c1p = PREVIOUS(c1);
			long c1n = NEXT(c1);
			long c2p = PREVIOUS(c2);
			long c2n = NEXT(c2);

			// if we found the match
			if(	(_triangleTable[c1] != _triangleTable[c2]) && 
				(_triangleTable[c1p] == _triangleTable[c2n]) &&
				(_triangleTable[c1n] == _triangleTable[c2p]))

				//(_triangleTable[PREVIOUS(c1)] == _triangleTable[NEXT(c2)]) &&
				//(_triangleTable[NEXT(c1)] == _triangleTable[PREVIOUS(c2)]))
			{
				_oppositeTable[c1] = c2;
				_oppositeTable[c2] = c1;
				break;
			}
			else if(j == triplets.size() -1 && _triangleTable[c1] == _triangleTable[c2])
			{
				bNoMatch = true;
				break;
			}
			// here, we will actually skip the last test if we don't put
			// j == triplets.size() -1

			// if we can't find the match in the current bin, then skip
			if(	triplets[i]._t1 != triplets[j]._t1 || 
				triplets[i]._t2 != triplets[j]._t2 )
				//||_triangleTable[c1] == _triangleTable[c2])
			{
				bNoMatch = true;
				// cout << " NO MATCH ";
				break;
			}
		}

		// if no match, then this may be the boundary triangle of a shell
		// test for this case

		if(bNoMatch)
		{	
			c2 = triplets[i+1]._t3;
			_oppositeTable[c1] = c2;
			_oppositeTable[c2] = c1;
			bNoMatch = false;
		}
	}

	cout << "opposite table size : " << (long)_oppositeTable.size() << endl;
	return true;
}

/**
 * constructOppositeTableBruteForce
 * O(n^2)
 * 
 **/
bool CTriangleMesh::buildOppositeTableBruteForce(void)
{
	long vtSize = (long)_triangleTable.size();
	long a, b, an, ap, bn, bp;
	if (0 == vtSize)
		return false;

	// keep track of which corner is assigned already
	int* flags = new int[vtSize];
	// set all the flags to false to start with
	memset(flags, 0, vtSize * sizeof(int)); 


	// initialize the opposite table using triangle table since
	// they have the same length
	_oppositeTable = _triangleTable;

	for(a = 0; a < vtSize; a++)
	{
		// already encounted the corner before
		if(flags[a] == 1)
			continue;

		for(b = 0; b < vtSize; b++)
		{
			// if already encounted the corner before or if we are testing
			// the same corner, skip
			if(flags[b] == 1 ||  _triangleTable[a] == _triangleTable[b])
				continue;
			
			an = NEXT(a);
			ap = PREVIOUS(a);
		
			bn = NEXT(b);
			bp = PREVIOUS(b);

			if(	_triangleTable[an] == _triangleTable[bp] && 
				_triangleTable[bn] == _triangleTable[ap])
			{
				_oppositeTable[a] = b;
				_oppositeTable[b] = a;
				flags[a] = 1;
				flags[b] = 1;
			}
		}
	}

	if(NULL != flags)
		delete flags;
	
	cout << "opposite table size : " << (long)_oppositeTable.size() << endl;
	return true;
}


bool CTriangleMesh::makeManifold()
{
	CPoint3D v2;
	long b, newv;
	
	long vSize = (long)_geometryTable.size();
	long cSize = (long)_triangleTable.size();

	bool * bVertexVisited = new bool[vSize];
	bool * bCornerVisited = new bool[cSize];
	memset(bVertexVisited, 0, sizeof(bool)*vSize);
	memset(bCornerVisited, 0, sizeof(bool)*cSize);

	// displayCornerTable();

	// int counter = 0;
	// int counter1 = 0;
	// cout  << "size triangleTable : " << cSize << endl;
	// cout << " cornerVisited at 220 is : " << bCornerVisited[57] << endl;
	for(long c = 0; c < (long)cSize; c++)
	{
		//cout << "corners : " << c << endl;
		if(bCornerVisited[c] == false)
		{
			// counter++;
			if(bVertexVisited[_triangleTable[c]] == 1)
			{
			//	cout << "corners : " << c << endl;

				v2 = CPoint3D(_geometryTable[_triangleTable[c]]);
				_geometryTable.push_back(v2);
				
				newv = (long)_geometryTable.size()-1;
				// cout << newv << endl;
				_triangleTable[c] = newv;
				bCornerVisited[c] = true;
				b = NEXT(_oppositeTable[NEXT(c)]);
				do{
					// cout << "b is : " << b << endl;
					_triangleTable[b] = newv;
					// cout << newv << endl;
					bCornerVisited[b] = true;
					b = NEXT(_oppositeTable[NEXT(b)]);
				} while(b != c);
			} // end if vertex visited
			else // vertex not visited
			{
				// counter1 ++;
				bVertexVisited[_triangleTable[c]] = true;
				bCornerVisited[c] = true;
				b = NEXT(_oppositeTable[NEXT(c)]);
				do{
					// cout << "b is : " << b << endl;
					bCornerVisited[b] = true;
					b = NEXT(_oppositeTable[NEXT(b)]);
				} while(b != c);
			} // end else vertex not visited
		} // end if corner not visited
	} // end foreach corner
	// cout << "went inside corner not visited : " << counter << endl;
	// cout << "went inside vertex not visited 1: " << counter1 << endl;

	return true;
}


/**
 * relaxation of points on a shpere
 * O(n^2) algorithm
 *
 * @param radius of influence, outside this radius would be considered
 *		  to have no influence over the current point
 * @param k small constant
 **/
void CTriangleMesh::relaxVerticesOnSphere(CPoint3D sphereCenter, double sphereRadius, double radius, double k)
{
	unsigned int i, j;

	CVector3 qp, repulsion, repulsionSum;
	CPoint3D p, p1, p2, q;

	for(i = 0; i < _geometryTable.size(); i++)
	{
		repulsionSum = CVector3(0, 0, 0);
		p = _geometryTable[i];
		for(j = 0; j < _geometryTable.size(); j++)
		{
			q = _geometryTable[j];

			// if q is the same point or q is located 
			// too far away from p, discard the point
			if((p-q).magnitude() > radius || p == q)
				continue;
			
			qp = p - q;
			// the repulsion force is inversely related to the
			// distance.
			repulsion = (p-q).normalized() * (radius/qp.magnitude());
			repulsionSum += repulsion;
		}

		// new position not on the sphere
		p1 = p + repulsionSum * k;

		// put p1 back onto the sphere;
		p2 = sphereCenter + (p1 - sphereCenter).normalized() * sphereRadius;
		_geometryTable[i] = p2;
	}
}

/**
 * relaxation of points on a shpere
 * O(n^2) algorithm
 *
 * @param radius of influence, outside this radius would be considered
 *		  to have no influence over the current point
 * @param k small constant
 **/
void CTriangleMesh::relaxVerticesInBall(CPoint3D ballCenter, double ballRadius, double radius, double k)
{
	unsigned int i, j;

	CVector3 qp, repulsion, repulsionSum;
	CPoint3D p, p1, p2, q;

	for(i = 0; i < _geometryTable.size(); i++)
	{
		repulsionSum = CVector3(0, 0, 0);
		p = _geometryTable[i];
		for(j = 0; j < _geometryTable.size(); j++)
		{
			q = _geometryTable[j];

			// if q is the same point or q is located 
			// too far away from p, discard the point
			if((p-q).magnitude() > radius || p == q)
				continue;
			
			qp = p - q;
			// the repulsion force is inversely related to the
			// distance.
			repulsion = (p-q).normalized() * (radius/qp.magnitude());
			repulsionSum += repulsion;
		}

		// new position not limited to be inside the ball
		p1 = p + repulsionSum * k;

		// if p1 goes outside the ball, put it back into the ball
		if((p1 - ballCenter).magnitude() > ballRadius) 
		{
			p2 = ballCenter + (p1 - ballCenter).normalized() * ballRadius;
			_geometryTable[i] = p2;
		}
		else
		{
			_geometryTable[i] = p1;
		}		
	}
}

/**
 * brute force generation
 *
 * O(n^4)
 *
 **/
bool CTriangleMesh::generateBruteForce(double radius)
{
	// check if this is an empty mesh
	if(0 == _geometryTable.size())
		return false;

	// first clear the triangle tables since we are generating them
	_triangleTable.clear();

	long i, j, k, l, newi, newj, newk, n=0;

	vector<CPoint3D> tempGeometryTable;

	//vector<CTriangle> triangles;
	//CTriangle triangle;

	double r2 = radius * 2;
	CPoint3D pi, pj, pk, pl;
	CPoint3D sphereCenterPos, sphereCenterNeg; 
	bool spherePosIsGood = true;
	bool sphereNegIsGood = true;
	
	long gtSize = (long)getVertexCount();
	
	// keep track of which vertices is used already
	int* flags = new int[gtSize];
	// set all the flags to false to start with
	memset(flags, -1, gtSize * sizeof(int)); 

	for(i = 0; i < gtSize; i++) 
	{
		cout << ".";
		for(j = i+1; j < gtSize; j++) 
		{
			for(k = j+1; k < gtSize; k++) 
			{
				pi = _geometryTable[i];
				pj = _geometryTable[j];
				pk = _geometryTable[k];

				// check if |jk| > 2r or |ik| > 2r, in these case, there can't be a sphere
				if ((pi-pj).magnitude() > r2 || (pj-pk).magnitude() > r2 || (pi-pk).magnitude() > r2)
					continue;

				// if the these three points can't have a common sphere, replace the third point and try again
				if(!sphereCenters(pi, pj, pk, radius, sphereCenterPos, sphereCenterNeg))
				{
					continue;
				}

				// test for the positive sphere center
				for(l = 0; l < gtSize; l++) 
				{
					if(l!=i && l!= j && l!=k) 
					{
						pl = _geometryTable[l];

						// test for positive center
						if((sphereCenterPos - pl).magnitude() < radius) 
							spherePosIsGood = false;
						
						// test for negative center
						if((sphereCenterNeg - pl).magnitude() < radius)
							sphereNegIsGood = false;
					}
				}

				if(spherePosIsGood) 
				{
					if (-1 == flags[i])
					{
						tempGeometryTable.push_back(pi);
						flags[i] = n;	// mark it
						newi = n++;
					}
					else	// already have it
					{
						newi = flags[i];
					}
					if (-1 == flags[j])
					{
						tempGeometryTable.push_back(pj);
						flags[j] = n;
						newj = n++;
					}
					else	// already have it
					{
						newj = flags[j];
					}
					if (-1 == flags[k])
					{
						tempGeometryTable.push_back(pk);
						flags[k] = n;
						newk = n++;
					}
					else	// already have it
					{
						newk = flags[k];
					}
					if(orientationTest(sphereCenterPos, pi, pj, pk)) 
					{
						_triangleTable.push_back(newi);
						_triangleTable.push_back(newj);
						_triangleTable.push_back(newk);
						//triangles.push_back(triangle);
					} else 
					{
						_triangleTable.push_back(newi);
						_triangleTable.push_back(newk);
						_triangleTable.push_back(newj);
						//triangles.push_back(triangle);
					}
				}
				if(sphereNegIsGood) 
				{
					if (-1 == flags[i])
					{
						tempGeometryTable.push_back(pi);
						flags[i] = n;	// mark it
						newi = n++;
					}
					else	// already have it
					{
						newi = flags[i];
					}
					if (-1 == flags[j])
					{
						tempGeometryTable.push_back(pj);
						flags[j] = n;
						newj = n++;
					}
					else	// already have it
					{
						newj = flags[j];
					}
					if (-1 == flags[k])
					{
						tempGeometryTable.push_back(pk);
						flags[k] = n;
						newk = n++;
					}
					else	// already have it
					{
						newk = flags[k];
					}
					if(orientationTest(sphereCenterNeg, pi, pj, pk)) 
					{
						_triangleTable.push_back(newi);
						_triangleTable.push_back(newj);
						_triangleTable.push_back(newk);
						//triangles.push_back(triangle);
					} 
					else 
					{
						_triangleTable.push_back(newi);
						_triangleTable.push_back(newk);
						_triangleTable.push_back(newj);
						//triangles.push_back(triangle);
					}
				}

				spherePosIsGood = true;
				sphereNegIsGood = true;
			}
		}
	}

	if (NULL != flags)
		delete flags;
	
	//displayVector(triangles);
	// remove all the redundant vertices
	_geometryTable = tempGeometryTable;
	cout << "# of points left : " << getVertexCount() << endl;
	cout << "# of triangles constructed : " << getTriangleCount() << endl;
	return true;
}




/**
 * traverseRecursive
 * traverse the triangle mesh recursively 
 */
void CTriangleMesh::traverseRecursive(long cornerIndex, CShell& shell)
{
	long triangleIndex = cornerIndex/3;

	long c = cornerIndex;
	long cn = NEXT(cornerIndex);
	long cp = PREVIOUS(cornerIndex);
	
	// if the current triangle doesn't belong to a shell yet
	if(-1 == _triangleShellIDTable[triangleIndex])
	{
		// shell._nTriangles ++;

		if(!_bVertexVisited[_triangleTable[c]])
		{
			shell._nVertices++;
			_bVertexVisited[_triangleTable[c]] = true;
		}
		if(!_bVertexVisited[_triangleTable[cn]])
		{
			shell._nVertices++;
			_bVertexVisited[_triangleTable[cn]] = true;
		}
		if(!_bVertexVisited[_triangleTable[cp]]){
			shell._nVertices++;
			_bVertexVisited[_triangleTable[cp]] = true;
		}

		_triangleShellIDTable[triangleIndex] = shell._id;
		//SET THE COLOR OF THE TRIANGLE TO THE COLOR OF THE CURRENT SHELL

		traverseRecursive(_oppositeTable[cn], shell);
		traverseRecursive(_oppositeTable[cp], shell);
	}
}

/**
 * traverse the triangle mesh in the edge breaker style
 *
 **/
void CTriangleMesh::traverseNonRecursive(long cornerIndex, CShell& shell) 
{
	cout << ".";
	long c = cornerIndex;

	// repeat step
	while(1) {
		long triangleIndex = c/3;
		long cn = NEXT(c);
		long cp = PREVIOUS(c);
		long cr = _oppositeTable[cn];
		long cl = _oppositeTable[cp];

		if(_triangleShellIDTable[triangleIndex] == -1) 
		{ // keep count on # of triangles
			// shell._nTriangles++;
			// add in the triangles
			shell._triangles.push_back(_triangleTable[cp]);
			shell._triangles.push_back(_triangleTable[c]);
			shell._triangles.push_back(_triangleTable[cn]);
		}

		if(!_bVertexVisited[_triangleTable[c]])
		{
			shell._nVertices++;
			_bVertexVisited[_triangleTable[c]] = true;
		}
		if(!_bVertexVisited[_triangleTable[cn]])
		{
			shell._nVertices++;
			_bVertexVisited[_triangleTable[cn]] = true;
		}
		if(!_bVertexVisited[_triangleTable[cp]]){
			shell._nVertices++;
			_bVertexVisited[_triangleTable[cp]] = true;
		}

		_triangleShellIDTable[triangleIndex] = shell._id; // may have to change to shell id, LATER
		if(_bVertexVisited[_triangleTable[c]] == false) 
		{
			_bVertexVisited[_triangleTable[c]] = true;
			c = cr; // c = c.r, since c.r = (c.n).o
		} 
		else if(_triangleShellIDTable[cr/3] != -1) 
		{ // if(c.r.t.m == 1)
			if(_triangleShellIDTable[cl/3] != -1) 
			{ // if(c.l.t.m == 1)
				return; // we've encountered the E case
			} 
			else 
			{
				c = cl;
			}
		} 
		else if(_triangleShellIDTable[cl/3] != -1) 
		{ // if(c.l.t.m) == 1
			c = cr;
		} 
		else 
		{
			traverseNonRecursive(cr, shell);
			c = cl;
		}
	}
}


void CTriangleMesh::traverse()
{
	long tSize = (long)getTriangleCount(); 

	// if no triangles to traverse, or no opposite corner table constricted
	if(0 == _oppositeTable.size() || 0 == tSize)
		return;

	// if already traversed.
	if(0 != _shellTable.size())
		return;

	long i, j;

	// initialize the color table
	_triangleShellIDTable = vector<long>(tSize);
	_bVertexVisited = vector<bool>(getVertexCount());

	// the shell ID of each triangle is initially set to -1
	// before traversal
	for(i = 0; i < tSize ; i++)
		_triangleShellIDTable[i] = -1;
	
	// for each corner of each triangle
	for(i = 0; i < tSize*3; i+= 3)
	{
		for(j = 0; j < (long)getVertexCount(); j++)
			_bVertexVisited[j] = false;

		if(-1 == _triangleShellIDTable[i/3])
		{
			// numShells++;
			// int info[2]; // num triangles, num vertices
			
			CShell shell;

			shell._id = (long)_shellTable.size();
			shell._startingCorner = i;
			//traverseRecursive(i, shell);
			traverseNonRecursive(i, shell);
			_shellTable.push_back(shell);
			shell._nGenus = long(shell._triangles.size())/4 - shell._nVertices/2 + 1;

			// isplayVector(_shellTable);
			// displayVector(shell._triangles);
			/*
			cout	<< "Shell [" << getShellCount()-1 
					<< "] (#v : " << shell._nVertices
					<< " | #t : " << shell._nTriangles
					<< " | #h : " << shell._nGenus << endl;
					*/
		}
	}

	// empty the temporary holder of flags
	_bVertexVisited.clear();
	cout << "The mesh had " << getShellCount() <<" shells\n";
}



// ------------------------------------------------------
// Local helper functions 
// ------------------------------------------------------
bool CTriangleMesh::orientationTest(CPoint3D c, CPoint3D p, CPoint3D q, CPoint3D r) 
// bool orientationTest(CPoint3D c, CPoint3D p, CPoint3D q, CPoint3D r) 
{
	CVector3 pq = q - p;
	CVector3 pr = r - p;
	CVector3 pc = c - p;
	if(pc * pq.cross(pr) > 0)
		return true;
	return false;
}

double CTriangleMesh::twoLinEqs(double a, double b, double c, double d, double e, double f) 
// double twoLinEqs(double a, double b, double c, double d, double e, double f) 
{
	double temp = b*d-a*e;
	if(temp == 0)
		return LONG_MAX;
	return (c*d - a*f)/temp;
}

CPoint3D CTriangleMesh::circumscribingCenter(CPoint3D p, CPoint3D q, CPoint3D r) 
// CPoint3D circumscribingCenter(CPoint3D p, CPoint3D q, CPoint3D r) 
{
	CVector3 pq, pr, perp1, perp2, pqxpr;
	CPoint3D m1, m2, center;

	double gamma;

	pq = q - p;
	pr = r - p;

	pqxpr = pq.cross(pr);
	perp1 = pq.cross(pqxpr);
	perp2 = pr.cross(pqxpr);

	m1 = (p + q)/2.0;
	m2 = (p + r)/2.0;

	gamma = twoLinEqs(perp1[0], -perp2[0], m2[0] - m1[0], perp1[1], -perp2[1], m2[1] - m1[1]);

	center = m2 + perp2 * gamma;

	return center;
}

CPoint3D CTriangleMesh::sphereCenter(CPoint3D p, CPoint3D q, CPoint3D r, double radius, bool positiveSign) 
//CPoint3D sphereCenter(CPoint3D p, CPoint3D q, CPoint3D r, double radius, bool positiveSign) 
{
	CPoint3D circleCenter = circumscribingCenter(p, q, r);
	CPoint3D SphereCenter;

	CVector3 pq, pr;
	double circleRadius;
	pq = q - p;
	pr = r - p;
	circleRadius = (p - circleCenter).magnitude();

	if(circleRadius > radius)
		return CPoint3D(LONG_MAX, LONG_MAX, LONG_MAX);

	if(positiveSign) {
		SphereCenter = circleCenter + pq.cross(pr).normalized()*
			(sqrt(radius*radius - circleRadius*circleRadius));
	} else {
		SphereCenter = circleCenter - pq.cross(pr).normalized()*
			(sqrt(radius*radius - circleRadius*circleRadius));
	}
	return SphereCenter;
}

/**
 * sphereCenters
 * Calculate the 2 sphere centers that are decided by the three points p, q, r
 * 
 * @param p, q, r the 3 points to decide the sphere
 * @param radius the radius of the sphere
 * @param pos the sphere center with the positive sign
 * @param neg the sphere center with the negative sign
 * @return false if the three points can't coincide on a sphere with the given radius
 *
 **/
bool CTriangleMesh::sphereCenters(CPoint3D p, CPoint3D q, CPoint3D r, double radius, CPoint3D& pos, CPoint3D& neg) 
// bool sphereCenters(CPoint3D p, CPoint3D q, CPoint3D r, double radius, CPoint3D& pos, CPoint3D& neg) 
{
	CPoint3D circleCenter = circumscribingCenter(p, q, r);
	CPoint3D SphereCenter;

	CVector3 pq, pr, dir;
	double circleRadius, dist;
	pq = q - p;
	pr = r - p;
	circleRadius = (p - circleCenter).magnitude();

	if(circleRadius > radius)
		return false;

	dir = pq.cross(pr).normalized();
	dist = sqrt(radius*radius - circleRadius*circleRadius);

	pos = circleCenter + dir*dist;
	neg = circleCenter - dir*dist;

	return true;
}


/**
 * generateFast
 *
 * voxelize the grid and only check points in the neighbouring grid
 *
 * O(n * (k * 27)^3) where k is the average number of vertices in a cell grid
 * hopefully, k*27 is much smaller than n
 *
 **/
bool CTriangleMesh::generateFast(double radius)
{

	// check if this is an empty mesh
	if(0 == _geometryTable.size())
		return false;

	// first clear the triangle tables since we are generating them
	_triangleTable.clear();

	long cellIndex, ic, jc, kc, lc, i, j, k, l, newi, newj, newk, n = 0;
	vector<CPoint3D> tempGeometryTable;
	double r2 = radius * 2;

	CPoint3D sphereCenterPos, sphereCenterNeg, pi, pj, pk, pl;
	bool spherePosIsGood = true;
	bool sphereNegIsGood = true;
	
	long gtSize = (long)getVertexCount();
	
	// keep track of which vertices is used already
	int* flags = new int[gtSize];
	// set all the flags to false to start with
	memset(flags, -1, gtSize * sizeof(int)); 

	// store the vertex id of the points in the neighbouring cells
	vector<long> neighbours;
	//vector<CTriangle> triangles;
	//CTriangle triangle;

	// whether this ordered set of vertices already tested
	long index = 0;
	long size = gtSize*gtSize*gtSize;
	bool* verticesTested = new bool[size];
	if(NULL == verticesTested)
	{
		cout << "+++++++++ NOT ENOUGH MEMORY +++++++++++" << endl;
		exit(-1);
	}
	// set to false first
	memset(verticesTested, 0, size*sizeof(bool));


	// get the min-max box and store the two corners
	// of the box into _min and _max
	setMinMaxBox();
	cout << "min : " << _min << endl;
	cout << "max : " << _max << endl;


	// according to the min-max box, quantize the space containing
	// the entire triangle mesh into sqaure cells of side radius*2
	quantize(2*radius);

	// loop through all quantized cells to get all points
	for(cellIndex = 0; cellIndex < (long)_cells.size(); cellIndex++) 
	{
		cout << ".";

		// if this cell is empty, go to next cell
		if(0 == _cells[cellIndex].size()) 
			continue;

		// cout << "cell (" << cellIndex << ") " << endl;
		// displayVector(_cells[cellIndex]);

		// store all possible points contained in this cell
		// and in the 26 surrounding cells
		neighbours = getNeighbours(cellIndex);

		// displayVector(neighbours);

		// for all points in this cell
		for(ic = 0; ic < (long)_cells[cellIndex].size(); ic++)
		{
			// get the coordinates of the first point
			i = _cells[cellIndex][ic];

			for(jc = 0; jc < (long)neighbours.size(); jc++)
			{
				// get the coordinates of the second point
				j = neighbours[jc];
				// check if the first point is the same as the second point
				if (i == j) continue;

				pi = _geometryTable[i];
				pj = _geometryTable[j];
				// check if |ab| > 2r, in that case, there can't be a sphere
				if ((pi-pj).magnitude() > r2)
				{
					continue;
				}

				for(kc = jc+1; kc < (long)neighbours.size(); kc++)
				{
					k = neighbours[kc];
					if(i == k || j == k) continue;

					// HOWARD: here, when we examine the neighbours of the current cell, we might encounter
					// triangles that are already constructed when we check one of the neighbouring cells
					// we want to be able to not add in those triangles, then we need some kind of predicate
					// that the triangle with these three vertices has been constructed before.
					index = hash(i, j, k);

					// if this vertex set is already tested, skip, otherwise, mark it.
					if(verticesTested[index]) 
						continue;
					verticesTested[index] = true;

					pk = _geometryTable[k];

					// check if |jk| > 2r or |ik| > 2r, in these case, there can't be a sphere
					if ((pj-pk).magnitude() > r2 || (pi-pk).magnitude() > r2)
						continue;

					// if the these three points can't have a common sphere, replace the third point and try again
					if(!sphereCenters(pi, pj, pk, radius, sphereCenterPos, sphereCenterNeg))
						continue;

					// test for sphere
					for(lc = 0; lc < (long)neighbours.size(); lc++) 
					{
						l = neighbours[lc];
						if(l!=i && l!= j && l!=k) 
						{
							pl = _geometryTable[l];

							// test for positive center
							if((sphereCenterPos - pl).magnitude() < radius) 
								spherePosIsGood = false;
							
							// test for negative center
							if((sphereCenterNeg - pl).magnitude() < radius)
								sphereNegIsGood = false;
						}
					}

					if(spherePosIsGood) 
					{
						if (-1 == flags[i])
						{
							tempGeometryTable.push_back(pi);
							flags[i] = n;	// mark it
							newi = n++;
						}
						else	// already have it
						{
							newi = flags[i];
						}
						if (-1 == flags[j])
						{
							tempGeometryTable.push_back(pj);
							flags[j] = n;
							newj = n++;
						}
						else	// already have it
						{
							newj = flags[j];
						}
						if (-1 == flags[k])
						{
							tempGeometryTable.push_back(pk);
							flags[k] = n;
							newk = n++;
						}
						else	// already have it
						{
							newk = flags[k];
						}
						_triangleTable.push_back(newi);
						_triangleTable.push_back(newj);
						_triangleTable.push_back(newk);

					}
					if(sphereNegIsGood) 
					{
						if (-1 == flags[i])
						{
							tempGeometryTable.push_back(pi);
							flags[i] = n;	// mark it
							newi = n++;
						}
						else	// already have it
						{
							newi = flags[i];
						}
						if (-1 == flags[j])
						{
							tempGeometryTable.push_back(pj);
							flags[j] = n;
							newj = n++;
						}
						else	// already have it
						{
							newj = flags[j];
						}
						if (-1 == flags[k])
						{
							tempGeometryTable.push_back(pk);
							flags[k] = n;
							newk = n++;
						}
						else	// already have it
						{
							newk = flags[k];
						}
						_triangleTable.push_back(newi);
						_triangleTable.push_back(newk);
						_triangleTable.push_back(newj);
					}

					spherePosIsGood = true;
					sphereNegIsGood = true;
				}
			}
		}
	}
	// displayVector(triangles);

	// clean up the cells
	_cells.clear();
	
	// release memory
	if(NULL != verticesTested)
		delete verticesTested;
	
	if(NULL != flags)
		delete flags;

	// remove all the redundant vertices
	_geometryTable = tempGeometryTable;
	cout << "# of points left : " << getVertexCount() << endl;
	cout << "# of triangles constructed : " << getTriangleCount() << endl;
	return true;
}

/*
bool CTriangleMesh::generateFast(double radius)
{

	// check if this is an empty mesh
	if(0 == _geometryTable.size())
		return false;

	// first clear the triangle tables since we are generating them
	_triangleTable.clear();

	long cellIndex, ic, jc, kc, lc, i, j, k, l, newi, newj, newk, nP=0, nN=0;
	vector<CPoint3D> tempGeometryTableP;
	vector<CPoint3D> tempGeometryTableN;
	vector<long>	 tempTriangleTableN;

	CPoint3D sphereCenterPos, sphereCenterNeg, pi, pj, pk, pl;
	bool spherePosIsGood = true;
	bool sphereNegIsGood = true;
	double r2 = radius * 2;

	long gtSize = (long)getVertexCount();
	
	// keep track of which vertices is used already
	int* flagsP = new int[gtSize];
	// set all the flags to false to start with
	memset(flagsP, -1, gtSize * sizeof(int)); 

	// keep track of which vertices is used already
	int* flagsN = new int[gtSize];
	// set all the flags to false to start with
	memset(flagsN, -1, gtSize * sizeof(int)); 

	// store the vertex id of the points in the neighbouring cells
	vector<long> neighbours;
	//vector<CTriangle> triangles;
	//CTriangle triangle;

	// whether this ordered set of vertices already tested
	long index = 0;
	long size = gtSize*gtSize*gtSize;
	bool* verticesTested = new bool[size];
	if(NULL == verticesTested)
	{
		cout << "+++++++++ NOT ENOUGH MEMORY +++++++++++" << endl;
		exit(-1);
	}
	// set to false first
	memset(verticesTested, 0, size*sizeof(bool));


	// get the min-max box and store the two corners
	// of the box into _min and _max
	setMinMaxBox();
	cout << "min : " << _min << endl;
	cout << "max : " << _max << endl;


	// according to the min-max box, quantize the space containing
	// the entire triangle mesh into sqaure cells of side radius*2
	quantize(2*radius);

	// loop through all quantized cells to get all points
	for(cellIndex = 0; cellIndex < (long)_cells.size(); cellIndex++) 
	{
		cout << ".";

		// if this cell is empty, go to next cell
		if(0 == _cells[cellIndex].size()) 
			continue;

		// cout << "cell (" << cellIndex << ") " << endl;
		// displayVector(_cells[cellIndex]);

		// store all possible points contained in this cell
		// and in the 26 surrounding cells
		neighbours = getNeighbours(cellIndex);

		// displayVector(neighbours);

		// for all points in this cell
		for(ic = 0; ic < (long)_cells[cellIndex].size(); ic++)
		{
			// get the coordinates of the first point
			i = _cells[cellIndex][ic];

			for(jc = 0; jc < (long)neighbours.size(); jc++)
			{
				// get the coordinates of the second point
				j = neighbours[jc];
				// check if the first point is the same as the second point
				if (i == j) continue;

				pi = _geometryTable[i];
				pj = _geometryTable[j];
				// check if |ab| > 2r, in that case, there can't be a sphere
				if ((pi-pj).magnitude() > r2)
				{
					continue;
				}


				for(kc = jc+1; kc < (long)neighbours.size(); kc++)
				{
					k = neighbours[kc];
					if(i == k || j == k) continue;

					// HOWARD: here, when we examine the neighbours of the current cell, we might encounter
					// triangles that are already constructed when we check one of the neighbouring cells
					// we want to be able to not add in those triangles, then we need some kind of predicate
					// that the triangle with these three vertices has been constructed before.
					index = hash(i, j, k);

					// if this vertex set is already tested, skip, otherwise, mark it.
					if(verticesTested[index]) 
						continue;
					verticesTested[index] = true;

					pk = _geometryTable[k];

					// check if |jk| > 2r or |ik| > 2r, in these case, there can't be a sphere
					if ((pj-pk).magnitude() > r2 || (pi-pk).magnitude() > r2)
						continue;

					// if the these three points can't have a common sphere, replace the third point and try again
					if(!sphereCenters(pi, pj, pk, radius, sphereCenterPos, sphereCenterNeg))
						continue;

					// test for sphere
					for(lc = 0; lc < (long)neighbours.size(); lc++) 
					{
						l = neighbours[lc];
						if(l!=i && l!= j && l!=k) 
						{
							pl = _geometryTable[l];

							// test for positive center
							if((sphereCenterPos - pl).magnitude() < radius) 
								spherePosIsGood = false;
							
							// test for negative center
							if((sphereCenterNeg - pl).magnitude() < radius)
								sphereNegIsGood = false;
						}
					}

					if(spherePosIsGood) 
					{
						if (-1 == flagsP[i])
						{
							tempGeometryTableP.push_back(pi);
							flagsP[i] = nP;	// mark it
							newi = nP++;
						}
						else	// already have it
						{
							newi = flagsP[i];
						}
						if (-1 == flagsP[j])
						{
							tempGeometryTableP.push_back(pj);
							flagsP[j] = nP;
							newj = nP++;
						}
						else	// already have it
						{
							newj = flagsP[j];
						}
						if (-1 == flagsP[k])
						{
							tempGeometryTableP.push_back(pk);
							flagsP[k] = nP;
							newk = nP++;
						}
						else	// already have it
						{
							newk = flagsP[k];
						}
						_triangleTable.push_back(newi);
						_triangleTable.push_back(newj);
						_triangleTable.push_back(newk);

					}
					if(sphereNegIsGood) 
					{
						if (-1 == flagsN[i])
						{
							tempGeometryTableN.push_back(pi);
							flagsN[i] = nN;	// mark it
							newi = nN++;
						}
						else	// already have it
						{
							newi = flagsN[i];
						}
						if (-1 == flagsN[j])
						{
							tempGeometryTableN.push_back(pj);
							flagsN[j] = nN;
							newj = nN++;
						}
						else	// already have it
						{
							newj = flagsN[j];
						}
						if (-1 == flagsN[k])
						{
							tempGeometryTableN.push_back(pk);
							flagsN[k] = nN;
							newk = nN++;
						}
						else	// already have it
						{
							newk = flagsN[k];
						}
						tempTriangleTableN.push_back(newi);
						tempTriangleTableN.push_back(newk);
						tempTriangleTableN.push_back(newj);
					}

					spherePosIsGood = true;
					sphereNegIsGood = true;
				}
			}
		}
	}
	// displayVector(triangles);

	// clean up the cells
	_cells.clear();
	
	// release memory
	if(NULL != verticesTested)
		delete verticesTested;

	cout << "\ngtp : " << tempGeometryTableP.size() << endl;
	cout << "gtn : " << tempGeometryTableN.size() << endl;
	cout << "ttp : " << _triangleTable.size() << endl;
	cout << "ttn : " << tempTriangleTableN.size() << endl;

	// remove all the redundant vertices
	_geometryTable = tempGeometryTableP;
	// combine positive and negative vertex table

	_geometryTable.insert(_geometryTable.end(), tempGeometryTableN.begin(), tempGeometryTableN.end());

	long psize = (long) tempGeometryTableP.size();
	// readjust the corner indexes
	for(i = 0; i < (long)tempTriangleTableN.size(); i++)
		tempTriangleTableN[i] += psize;

	// combine positive and negative vertex table
	_triangleTable.insert(_triangleTable.end(), tempTriangleTableN.begin(), tempTriangleTableN.end());

	// displayVector(_triangleTable);

	cout << "# of points left : " << getVertexCount() << endl;
	cout << "# of triangles constructed : " << getTriangleCount() << endl;
	return true;
}
*/

/**
 * hash to come up an index
 **/
long CTriangleMesh::hash(long v1, long v2, long v3)
{
	long n = (long) _geometryTable.size();
	long min, mid, max;
	// sort the vertices in the assending order
	if(v1 <= v2 && v1 <= v3)
	{	
		min = v1;
		if(v2<v3)
		{
			mid = v2;
			max = v3;
		}
		else
		{
			mid = v3;
			max = v2;
		}
	}
	else if(v2 <= v1 && v2 <= v3)
	{	
		min = v2;
		if(v1<v3)
		{
			mid = v1;
			max = v3;
		}
		else
		{
			mid = v3;
			max = v1;
		}
	}
	else if(v3 <= v1 && v3 <= v2)
	{	
		min = v3;
		if(v2<v1)
		{
			mid = v2;
			max = v1;
		}
		else
		{
			mid = v1;
			max = v2;
		}
	}
	return (min*n*n + mid*n + max);
}


/**
 * getCellCoordinates
 *
 * get the cell coordinates in the grid from the cell index number
 **/
inline
void CTriangleMesh::getCellCoordinates(long cellIndex, long* pxi, long* pyi, long* pzi)
{
	long xd = _grid._xDim;
	long yd = _grid._yDim;

	// based on the dimensions, figure out the x, y, z coordinate of the current cell
	*pzi = cellIndex/(xd*yd);
	long yr = cellIndex%(xd*yd);
	*pyi = yr/xd;
	*pxi = yr%xd;
}

inline
void CTriangleMesh::getCellCoordinates(CPoint3D p, long *pxi, long *pyi, long* pzi)
{
	// calculate the cell this point will be in
	*pxi = (long)floor((p[0] - _min[0]) * _grid._xFrac);
	*pyi = (long)floor((p[1] - _min[1]) * _grid._yFrac);
	*pzi = (long)floor((p[2] - _min[2]) * _grid._zFrac);
}


/**
 * get the min-max box and store the two corners
 * of the box into _min and _max
 *
 **/
void CTriangleMesh::setMinMaxBox(void)
{
	double minX = LONG_MAX;
	double minY = LONG_MAX;
	double minZ = LONG_MAX;
	double maxX = -LONG_MAX;
	double maxY = -LONG_MAX;
	double maxZ = -LONG_MAX;
	vector<CPoint3D>::iterator curr;
	CPoint3D p;

	for(curr = _geometryTable.begin();
		curr != _geometryTable.end();
		curr ++)
	{
		p = *curr;
		if(minX > p[0])
			minX = p[0];
		if(maxX < p[0])
			maxX = p[0];
		if(minY > p[1])
			minY = p[1];
		if(maxY < p[1])
			maxY = p[1];
		if(minZ > p[2])
			minZ = p[2];
		if(maxZ < p[2])
			maxZ = p[2];
	}

	_min = CPoint3D(minX, minY, minZ);
	_max = CPoint3D(maxX, maxY, maxZ);
}


/**
 * according to the min-max box, quantize the space containing
 * the entire triangle mesh into sqaure cells of side radius*2
 **/
void CTriangleMesh::quantize(double side)
{
	long xd, yd, zd, i;
	long cellX, cellY, cellZ;
	double rx, ry, rz;
	CPoint3D p;

	// get the dimension of the quantized cell grid
	xd = (int)ceil((_max[0] - _min[0])/side);
	yd = (int)ceil((_max[1] - _min[1])/side);
	zd = (int)ceil((_max[2] - _min[2])/side);
	
	cout << "# of cells : <" << xd << ", " << yd << ", " << zd << ">\n";

	// create the cells
	_cells = vector<longList>(xd * yd * zd);

	// calculate the ratio
	rx = xd/ceil(_max[0] - _min[0]);
	ry = yd/ceil(_max[1] - _min[1]);
	rz = zd/ceil(_max[2] - _min[2]);

	// save the information
	_grid._xDim = xd;
	_grid._yDim = yd;
	_grid._zDim = zd;
	_grid._xFrac = rx;
	_grid._yFrac = ry;
	_grid._zFrac = rz;

	// assign each vertex into cells
	for(i = 0; i < (long)_geometryTable.size(); i++) 
	{
		p = _geometryTable[i]; 
		// calculate the cell this point will be in
		cellX = (long)floor((p[0] - _min[0]) * rx);
		cellY = (long)floor((p[1] - _min[1]) * ry);
		cellZ = (long)floor((p[2] - _min[2]) * rz);

		_cells[xd*yd*cellZ + xd*cellY + cellX].push_back(i);
	}
}


vector<long> CTriangleMesh::getNeighbours(long cellIndex) 
{
	vector<long> neighbours;
	long neighbourIndex;

	long xi, yi, zi;

	// based on the dimensions, figure out the x, y, z coordinate of the current cell
	getCellCoordinates(cellIndex, &xi, &yi, &zi);

	long xd = _grid._xDim;
	long yd = _grid._yDim;
	long zd = _grid._zDim;

	// if the current cell is on the boundary, special treatment is needed
	long x0 = (xi <= 0)?0:xi-1;
	long x1 = (xi >= xd-1)?xd-1:xi+1;
	long y0 = (yi <= 0)?0:yi-1;
	long y1 = (yi >= yd-1)?yd-1:yi+1;
	long z0 = (zi <= 0)?0:zi-1;
	long z1 = (zi >= zd-1)?zd-1:zi+1;

	// accumulate the neighbouring points
	for(int i = x0; i <= x1; i++) 
	{
		for(int j = y0; j <= y1; j++) 
		{
			for(int k = z0; k <= z1; k++) 
			{
				neighbourIndex = xd * yd * k + xd * j + i;
				neighbours.insert(neighbours.end(), _cells[neighbourIndex].begin(), _cells[neighbourIndex].end());
			}
		}
	}
	return neighbours;
}

/**
 *
 * HOWARD: here, when we examine the neighbours of the current cell, we might encounter
 * triangles that are already constructed when we check one of the neighbouring cells
 * we want to be able to not add in those triangles, then we need some kind of predicate
 * that the triangle with these three vertices has been constructed before.
 *
 * since we traverse the cell in the top-down, left-right fashion, so if we encounter
 * a triangle (one vertex in the current cell for sure) with at least one vertex in the 
 * previous cell, and the other vertex in the cells that is the neighbour of the previous
 * cell, then this vertex configuration is tested already.
 *
 **/