/** =================================================================\n
 *
 * TriangleMesh Compression Project - CS 7491 2004 project 1\n
 * @file TriangleMesh.h
 *
 * Primary Authors: Brandon Beck
 *					Nguyen Truong (nltruong@cc.gatech.edu)\n
 *                  Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright 2004\n
 * (C) Copyright Georgia Institute of Technology, 2004\n
 *
 * ==================================================================\n
 * Basic TriangleMesh class definition
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2004-02-14
 * @author Brandon Beck, Nguyen Truong, Howard Zhou
 *
 *
 * @todo
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#ifndef __CTRIANGLEMESH_H__
#define __CTRIANGLEMESH_H__

#include <iostream>
#include <vector>
#include "..\MatrixLib\vector.h"
// #include "common.h"

/// @namespace using standard namespace
using namespace std;

class CShell
{
public:
	long _id;
	long _startingCorner;
	//long _nTriangles;
	long _nVertices;
	long _nGenus;
	vector<long> _triangles;

	CShell()
	{
		_id = -1;
		_startingCorner = -1;
		// _nTriangles = 0;
		_nVertices = 0;
		_nGenus = 0;
	}

	friend ostream& operator << (ostream& os, CShell & shell)
	{
		os	<< " shell(" 
			<< shell._id << ", " 
			<< shell._startingCorner << ", "
			<< shell._nVertices << ", "
			// << shell._nTriangles << ", "
			<< shell._nGenus << ") ";
			
		return os;
	}
};


class CGridInfo
{
public:
	// how many cells in this grid in each dimension
	long _xDim;
	long _yDim;
	long _zDim;
	// normalized unit size, treating the side of the grid as 1
	// what is the fraction of the side of a cell
	double _xFrac;	
	double _yFrac;
	double _zFrac;
};

/*
class CTriangle
{
private:
public:
	long _v1;
	long _v2;
	long _v3;

	CTriangle(){};

	CTriangle(long v1, long v2, long v3)
	{
		// sort the vertices in the assending order
		if(v1 <= v2 && v1 <= v3)
		{	
			_v1 = v1;
			_v2 = (v2<v3)?v2:v3;
			_v3 = (v2<v3)?v3:v2;
			return;
		}
		if(v2 <= v1 && v2 <= v3)
		{	
			_v1 = v2;
			_v2 = (v1<v3)?v1:v3;
			_v3 = (v1<v3)?v3:v1;
			return;
		}
		if(v3 <= v1 && v3 <= v2)
		{	
			_v1 = v3;
			_v2 = (v1<v2)?v1:v2;
			_v3 = (v1<v2)?v2:v1;
			return;
		}
	}

	bool operator == (const CTriangle & reft) const
	{
		return (_v1 == reft._v1 && _v2 == reft._v2 && _v3 == reft._v3);
	}

	friend ostream& operator << (ostream & os, const CTriangle & reft)
	{
		os << "T<" << reft._v1 << ", " << reft._v2 << ", " << reft._v3 << "> ";
		return os;
	}
};
*/

typedef vector<long> longList;

// ==================================================================================
/**
 * @class CTriangleMesh
 * @brief General TriangleMesh class definition
 * @ingroup TriangleMesh
 *
 * @par requirements
 *
 * @par license
 * This code is absolutely free to use and modify. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 *
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2004-02-14
 * @author Brandon Beck, Nguyen Truong, Howard Zhou
 *
 * This class is a general triangleMesh class. This is an abstract class serving as 
 * the triangleMesh interface.
 * 
 * @todo
 * @bug
 * @warning
 */
// ================================================================================

class CTriangleMesh
{
protected:


private:
	
	// 5 primary tables
	vector<CPoint3D> _geometryTable;
	vector<long>	_triangleTable;
	vector<long>	_oppositeTable;
	vector<CShell>	_shellTable;
	vector<long>	_triangleShellIDTable;	///< stores shell id for each triangle

	// intermediate tables that will not be accessed from outside
	vector<bool>	_bVertexVisited;		///< used to store whether an vertex been visited or not
	vector<longList>	_cells;					///< list of cells containing list of points in the cell
	// vector<CPoint3D> _colorTable;

	// min-max points of the triangle mesh
	// these two points are the two corners of a smallest rectangular box
	// that contains the entire triangle mesh
	CPoint3D		_min;
	CPoint3D		_max;
	CGridInfo		_grid;


	// ------------------------------------------------------
	// private helpers
	// ------------------------------------------------------
	void traverseRecursive(long index, CShell& shell);
	void traverseNonRecursive(long cornerIndex, CShell& shell);

	void setMinMaxBox(void);
	void quantize(double side);
	void getCellCoordinates(long cellIndex, long* pxi, long* pyi, long* pzi);
	void getCellCoordinates(CPoint3D p, long *pxi, long *pyi, long* pzi);
	vector<long> getNeighbours(long cellIndex);
	long hash(long i, long j, long k);

	/// local helper function declaration
	bool orientationTest(CPoint3D c, CPoint3D p, CPoint3D q, CPoint3D r);
	double twoLinEqs(double a, double b, double c, double d, double e, double f);
	CPoint3D circumscribingCenter(CPoint3D p, CPoint3D q, CPoint3D r);
	CPoint3D sphereCenter(CPoint3D p, CPoint3D q, CPoint3D r, double radius, bool positiveSign);
	bool sphereCenters(CPoint3D p, CPoint3D q, CPoint3D r, double radius, CPoint3D& pos, CPoint3D& neg);

public:

	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------

	///< Default constructor. Creates an uninitialized triangle mesh
	CTriangleMesh (){};

	///< copy constructor
	CTriangleMesh (const CTriangleMesh& refmesh);


	CTriangleMesh (vector<CPoint3D> points){_geometryTable = points;}

	// Destructor
	~CTriangleMesh (){};

	// ------------------------------------------------------
	// Predicators
	// ------------------------------------------------------
	virtual	bool isValid()			const{return true;};
	virtual bool isInitialized()	const{return true;};

	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------
	long				getVertexCount	(void) const {return (long)_geometryTable.size();};
	long				getTriangleCount(void) const {return (long)_triangleTable.size()/3;};	///< each triangle has 3 vertices
	long				getShellCount	(void) const {return (long)_shellTable.size();};
	vector<CPoint3D>	getVertices		(void) const {return _geometryTable;};
	vector<long>		getTriangles	(void) const {return _triangleTable;};
	vector<CShell>		getShells		(void) const {return _shellTable;};
	vector<long>		getTriangleShellIDs(void) const {return _triangleShellIDTable;};
	// vector<CPoint3D>	getColors		(void) const {return _colorTable;};

	void setVertices(const vector<CPoint3D>& points)
	{
		clear();
		_geometryTable = points;
	};


	// ------------------------------------------------------
	// Overloaded Operators 
	// ------------------------------------------------------
	
	// seems assingment operator is defined for user defined class
	// if the default constructor is used
	// CTriangleMesh& operator= (const CTriangleMesh& refmesh);

	// ------------------------------------------------------
	// Input, Output, and Test
	// ------------------------------------------------------
	virtual void readFile			(const char* filename);
	virtual void writeFile			(const char* filename) const;
	virtual void readCompressedFile	(const char* filename) {};
	virtual void writeCompressedFile(const char* filename) const{};

	// virtual void draw		(void) const{};
	void test(void);
	void displayOppositeTable(void);	
	void displayCornerTable(void);

	// ------------------------------------------------------
	// Other operation
	// ------------------------------------------------------
	void clear(void);
	void addVertices(vector<CPoint3D> points);

	bool generateBruteForce(double radius);
	bool generateFast(double radius);
	
	bool buildOppositeTableBruteForce(void);
	bool buildOppositeTable(void);

	bool makeManifold(void);

	void relaxVerticesOnSphere(CPoint3D sphereCenter, double sphereRadius, double radius, double k = 0.1);
	void relaxVerticesInBall  (CPoint3D ballCenter,   double ballRadius,   double radius, double k = 0.1);

	// Brandon's addition
	// void buildOTable();
	void traverse();

	// ------------------------------------------------------
	// Friend functions
	// ------------------------------------------------------

};

/**
 * copy constructor
 *
 **/
inline CTriangleMesh::CTriangleMesh (const CTriangleMesh& refmesh)
{
	if(0 != refmesh._geometryTable.size())
		_geometryTable = refmesh._geometryTable;
	if(0 != refmesh._triangleTable.size())
		_triangleTable = refmesh._triangleTable;
	if(0 != refmesh._oppositeTable.size())
		_oppositeTable = refmesh._oppositeTable;
	if(0 != refmesh._shellTable.size())
		_shellTable = refmesh._shellTable;
	if(0 != refmesh._triangleShellIDTable.size())
		_triangleShellIDTable = refmesh._triangleShellIDTable;
	
	_min = refmesh._min;
	_max = refmesh._max;
	_grid = refmesh._grid;
}

inline void CTriangleMesh::clear (void)
{
	_geometryTable.clear();
	_triangleTable.clear();
	_oppositeTable.clear();
	_shellTable.clear();
	_triangleShellIDTable.clear();
}


/**
 * operator = (const CVector& vec) \n
 * Copies the values of another vector to the 
 * corresponding elements of this vector.
 * @param refvector 
 * @return reference to the current vector 
 **/
/*
inline CTriangleMesh& CTriangleMesh::operator= (const CTriangleMesh& refmesh) 
{
	_geometryTable = refmesh._geometryTable;
	_triangleTable = refmesh._triangleTable;
	return (*this);
}	
*/

#endif // __CTRIANGLEMESH_H__