/** =================================================================\n
 *
 * CS 7491 2004\n
 * @file MeshCompression.cpp
 *
 * Primary Authors: Brandon Beck
 *					Nguyen Truong (nltruong@cc.gatech.edu)\n
 *                  Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright 2004\n
 * (C) Copyright Georgia Institute of Technology, 2004\n
 *
 * ==================================================================\n
 * Test case for triangle mesh compression and simplification
 *
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2004-02-13
 * @author Brandon Beck, Nguyen Truong, Howard Zhou
 *
 *
 * @todo
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */



#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <iostream>


#include "MeshCompression.h"
#include "..\Viewer\OpenGLWnd.h"
#include "PointSet.h"


#define IN_OUT			0
#define POINTSET_DEBUG  1
#define POINTSET_DEBUG2 0
#define SINGLE_SPHERE	0
#define SINGLE_BALL		0
#define SPHERE_AND_BALL	0


#ifndef PI
#define PI 3.141592654
#endif

using namespace std;


// global variables
vector<CPoint3D> gColorPallet;
int dup = 0;

bool isIn(CPoint3D p, CPoint3D center, double radius);


void CMeshCompression::run(void)
{
	_CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );

	vector<CPoint3D> points;

	// construct the color pallet
	gColorPallet.push_back(CPoint3D(1, 0, 0));
	gColorPallet.push_back(CPoint3D(1, 1, 0));
	gColorPallet.push_back(CPoint3D(1, 0, 1));
	gColorPallet.push_back(CPoint3D(0, 1, 0));
	gColorPallet.push_back(CPoint3D(0, 1, 1));


   clock_t start, finish;
   double  duration;


	cout << "\nMesh Compression\n=========================\n";

#if IN_OUT == 1

	_triangleMesh.readFile("..\\data\\test1.ov");
	// _triangleMesh.test();
	
	// _triangleMesh.traverse();

	// set up the rendering function
	_pWndMain->setRenderFunc( & CMeshCompression::renderGLScene );

	_pWndMain->updateAllViews();

	// _triangleMesh.writeFile("..\\data\\m1\\cow1.ov");

#endif


#if POINTSET_DEBUG == 1

	// HOWARD : NOTE
	// here, we have to strictly follow the "addition first, then
	// subtraction rule, in order to construct since the 
	// addition and subtraction of CPointSet is not communitive
	CPointSet<CPoint3D> pointSet;

	for(int i = 1; i < 3; i += 2)
	{

	int nPoints = i*50;
	double avgDist = 0.7;
	
	/*
	pointSet = CPointSet<CPoint3D>(avgDist, SPHERE, CPoint3D(-3, -2, 0), 1.5, true) 
		+ CPointSet<CPoint3D>(avgDist, BALL, CPoint3D(2, 1, 0), 1.5, true)
		+ CPointSet<CPoint3D>(avgDist, BALL, CPoint3D(1, 0, 0), 2, true)
		- CPointSet<CPoint3D>(avgDist, BLOCKER, CPoint3D(1, 2, 0), 2)	
	- CPointSet<CPoint3D>(avgDist, BLOCKER,  CPoint3D(-2, -2, 1), 1.5);
	*/
	
	//pointSet = CPointSet<CPoint3D>(avgDist, SPHERE, CPoint3D(0, 0, 0), 3, true)
	//	- CPointSet<CPoint3D>(avgDist, BLOCKER, CPoint3D(1.5, 1, 0), 3);

	/*
	pointSet = CPointSet<CPoint3D>(avgDist, BALL, CPoint3D(0, 0, 0), 3, true)
		- CPointSet<CPoint3D>(avgDist, BLOCKER, CPoint3D(0, 0, 0), 1.5)
		- CPointSet<CPoint3D>(avgDist, BLOCKER, CPoint3D(0, 2, 0), 1.5)
		- CPointSet<CPoint3D>(avgDist, BLOCKER, CPoint3D(0, -2, 0), 1.5)
		- CPointSet<CPoint3D>(avgDist, BLOCKER, CPoint3D(0, 4, 0), 3.5)
		- CPointSet<CPoint3D>(avgDist, BLOCKER, CPoint3D(0, -4, 0), 3.5);
	*/

	//pointSet = CPointSet<CPoint3D>(avgDist, BALL, CPoint3D(0, 0, 0), 3, true);
	// pointSet.writeFile("..\\data\\test.txt");
	// pointSet.readFile("..\\data\\comb002.txt");

	dup = (long)pointSet.getPoints().size();
	cout << "# of points generated in sphere : " << dup << endl;
	_triangleMesh.addVertices(pointSet.getPoints());

	// set up the rendering function
	_pWndMain->setRenderFunc( & CMeshCompression::renderGLScene );

	_pWndMain->updateAllViews();


	// generate triangles (triangle table)
	// _triangleMesh.generateBruteForce(1);

	double r = 1.5*avgDist;
	//double r = 2.0/sqrt(float(i));

	/*
	start = clock();
	_triangleMesh.generateBruteForce(r);
	finish = clock();

	duration = (double)(finish - start) / CLOCKS_PER_SEC;
	cout << "Time used : " << duration << endl;
	*/

	start = clock();
	// _triangleMesh.generateBruteForce(r);
	_triangleMesh.generateFast(r);
	finish = clock();

	duration = (double)(finish - start) / CLOCKS_PER_SEC;
	cout << "Time used : " << duration << endl;
	}

	// _triangleMesh.buildOppositeTableBruteForce();
	_triangleMesh.buildOppositeTable();
	_triangleMesh.makeManifold();
	
	cout << "v new : " <<  _triangleMesh.getVertexCount() << endl;
	_triangleMesh.buildOppositeTable();

	//_triangleMesh.displayOppositeTable();


#endif

#if POINTSET_DEBUG2 == 1

	// HOWARD : NOTE
	// here, we have to strictly follow the "addition first, then
	// subtraction rule, in order to construct since the 
	// addition and subtraction of CPointSet is not communitive

	for(int i = 8; i < 11; i ++)
	{
	/*
	pointSet = CPointSet<CPoint3D>(20, SPHERE, CPoint3D(-3, -2, 0), 1.5) 
		+ CPointSet<CPoint3D>(20, BALL, CPoint3D(3, 2, 0), 1.5)
		// + CPointSet<CPoint3D>(20, BALL, CPoint3D(1, 0, 0), 2)
		// - CPointSet<CPoint3D>(0, BLOCKER, CPoint3D(1, 2, 0), 2)
		- CPointSet<CPoint3D>(0, BLOCKER,  CPoint3D(-2, -2, 1), 1.5);
	*/

	/*
	pointSet = CPointSet<CPoint3D>(50, SPHERE, CPoint3D(0, 0, 0), 3)
		- CPointSet<CPoint3D>(0, BLOCKER, CPoint3D(0, 1, 0), 3);
	*/
	int nPoints = i*50;
	CPointSet<CPoint3D> pointSet = CPointSet<CPoint3D>((long)nPoints, SPHERE, CPoint3D(0, 0, 0), 3, true);

	cout << "# of points generated in sphere : " << pointSet.getPoints().size() << endl;
	_triangleMesh.setVertices(pointSet.getPoints());

	// set up the rendering function
	_pWndMain->setRenderFunc( & CMeshCompression::renderGLScene );

	_pWndMain->updateAllViews();

	// generate triangles (triangle table)
	// _triangleMesh.generateBruteForce(1);

	double r = 2.0/sqrt(float(i));
	cout << "\nRadius : " << r << endl;

	start = clock();
	_triangleMesh.generateBruteForce(r);
	finish = clock();

	duration = (double)(finish - start) / CLOCKS_PER_SEC;
	cout << "Time used : " << duration << endl;

	start = clock();
	// _triangleMesh.generateFast(r);
	finish = clock();

	duration = (double)(finish - start) / CLOCKS_PER_SEC;
	cout << "Time used : " << duration << endl;
	}

	_triangleMesh.buildOppositeTable();

#endif


#if SINGLE_SPHERE == 1
	
	CPoint3D sphereCenter = CPoint3D(0, 0, 0);
	double sphereRadius = 3;
	long numPoints = 100;

	points = sphereWithBlocker(numPoints, sphereCenter, sphereRadius);

	cout << "# of points generated in sphere : " << points.size() << endl;
	_triangleMesh.addVertices(points);

	// set up the rendering function
	_pWndMain->setRenderFunc( & CMeshCompression::renderGLScene );

	_pWndMain->updateAllViews();
	
	double sphereArea = 4 * PI * sphereRadius * sphereRadius; 
	double radiusInfluence = 2 * sqrt(sphereArea / numPoints);

	double k = 0.2;
	for(int i = 0; i< 10; i++)
	{
		_triangleMesh.relaxVerticesOnSphere(sphereCenter, sphereRadius, radiusInfluence, k);

		_pWndMain->updateAllViews();
		k -= 0.015;
	}

	// generate triangles (triangle table)
	_triangleMesh.generateBruteForce(1.5);
	// _triangleMesh.generateFast(1.5);

#endif	

#if SPHERE_AND_BALL == 1

	CPoint3D sphereCenter = CPoint3D(-3, -2, 0);
	double sphereRadius = 1.5;
	CPoint3D ballCenter = CPoint3D(3, 2, 0);
	double ballRadius = 1.5;
	long numPoints = 20;

	// generate the sphere
	points = sphereWithBlocker(numPoints, sphereCenter, sphereRadius, CPoint3D(-2, -2, 1), 1.5);
	cout << "# of points generated in sphere : " << points.size() << endl;
	_triangleMesh.addVertices(points);

	_pWndMain->setRenderFunc( & CMeshCompression::renderGLScene );

	// generate the ball
	points = ballWithBlocker(numPoints, ballCenter, ballRadius); // , CPoint3D(0, 1, 1), 2);
	cout << "# of points generated in ball: " << points.size() << endl;
	_triangleMesh.addVertices(points);

	// set up the rendering function
	_pWndMain->setRenderFunc( & CMeshCompression::renderGLScene );

	// generate triangles (triangle table)
	_triangleMesh.generateBruteForce(1.5f);
	// _triangleMesh.generateFast(1.5f);
#endif



	// construct the opposite table
	// _triangleMesh.buildOppositeTableBruteForce();
	// _triangleMesh.displayOppositeTable();

	// _triangleMesh.buildOppositeTable();
	// _triangleMesh.buildOTable();
	// _triangleMesh.displayOppositeTable();
	
	_triangleMesh.traverse();

	_triangleMesh.writeFile("..\\data\\test.ov");

	_pWndMain->updateAllViews();
}

void CMeshCompression::renderGLScene(CTask * task)
{	
	CMeshCompression * mc = (CMeshCompression *)task;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// Clear Screen And Depth Buffer

	glPushMatrix();
	//glTranslatef(0.0f, 0.0f, -15.0f);
	
	mc->drawPoints(mc->_triangleMesh.getVertices(), CPoint3D(1.0, 1.0, 1.0));
	if(mc->_triangleMesh.getTriangleCount() != 0)
	{	
		// HOWARD
		// NOTE: here we need to pass the reference to the triangle instead of passing by value
		// otherwise, the triangle information doesn't get updated later when the structure is changed
		// This is a strange behaviour I guess have things to do with passing function pointers.
		// mc->drawTriangles(mc->_triangleMesh.getVertices(), mc->_triangleMesh.getTriangles(), mc->_triangleMesh.getColors());
		mc->drawTriangleMesh(mc->_triangleMesh);
		// mc->drawTrianglesHack(mc->_triangleMesh.getVertices(), mc->_triangleMesh.getTriangles(), CPoint3D(1, 0, 0), CPoint3D(0, 0, 0), 3);
	}
	glPopMatrix();
}

void CMeshCompression::drawPoints(vector<CPoint3D> points, CPoint3D color)
{

	/*
	GLfloat x, y, z;
	vector<CPoint3D>::iterator curr;

	glColor3f((GLfloat)color[0], (GLfloat)color[1], (GLfloat)color[2]);
	// glBegin(GL_POINTS);
	for(curr = points.begin();
		curr != points.end();
		curr++)
	{
		x = (*curr)[0];
		y = (*curr)[1];
		z = (*curr)[2];
		glTranslatef(x, y, z);

		// glVertex3f(x, y, z);

		glutSolidSphere(0.05f, 8, 8);
		glTranslatef(-x, -y, -z);
	}
	*/
	//glEnd();

	GLfloat x, y, z;
	long i;

	glColor3f((GLfloat)color[0], (GLfloat)color[1], (GLfloat)color[2]);
	//glBegin(GL_POINTS);
	for(i = 0; i < (long)points.size(); i++)
	{
		// glColor3f((GLfloat)color[0], (GLfloat)color[1], (GLfloat)color[2]);
		if(i >= dup)
		// if (i == 22)		
			glColor3f(0, 0, 1);

		x = points[i][0];
		y = points[i][1];
		z = points[i][2];
		glTranslatef(x, y, z);

		// glVertex3f(x, y, z);

		glutSolidSphere(0.05f, 8, 8);
		glTranslatef(-x, -y, -z);
	}
	// glEnd();

}

/**
 * drawTriangleMesh
 *
 * @param CTriangleMesh&
 *
 **/
void CMeshCompression::drawTriangleMesh(CTriangleMesh& triangleMesh)
{
	// draw the triangle meshes, assigne the color according to which shell it belongs.
	// drawTriangles(triangleMesh.getVertices(), triangleMesh.getTriangles(), 
	//		triangleMesh.getTriangleShellIDs(), gColorPallet);

	
	CPoint3D pointA, pointB, pointC, color;
	CVector3 normalVector;
	CShell shell;
	long i, j;
	int colorPalletSize = gColorPallet.size();
	vector<CPoint3D> vertices = triangleMesh.getVertices();
	vector<CShell> shells = triangleMesh.getShells();
	vector<long> triangles = triangleMesh.getTriangles();

	// if not traversed, use gray color
	if (0 == shells.size())
	{
		glColor3f(0.5, 0.5, 0.5);
		for(i = 0; i < (long)triangles.size(); i+= 3)
		{
			pointA = vertices[triangles[i]];
			pointB = vertices[triangles[i+1]];
			pointC = vertices[triangles[i+2]];

			normalVector = ((CVector3)(pointB - pointA)).cross(pointC - pointA);
			normalVector.normalize();

			glBegin(GL_TRIANGLES);
			{
				glNormal3f(normalVector[0], normalVector[1], normalVector[2]);
				glVertex3f(pointA[0], pointA[1], pointA[2]);
				glVertex3f(pointB[0], pointB[1], pointB[2]);
				glVertex3f(pointC[0], pointC[1], pointC[2]);
			}glEnd();
		}
	}
	else
	{
		// if traversed, draw each shell at a time
		for(i= 0; i < (long)shells.size(); i++)
		{
			shell = shells[i];

			// i is the current shell id
			// rotate the pallet to select a color
			color = gColorPallet[i % colorPalletSize];
			glColor3f((GLfloat)color[0], (GLfloat)color[1], (GLfloat)color[2]);

			for(j = 0; j < (long)shell._triangles.size(); j+= 3)
			{
				pointA = vertices[shell._triangles[j]];
				pointB = vertices[shell._triangles[j+1]];
				pointC = vertices[shell._triangles[j+2]];
			
				normalVector = ((CVector3)(pointB - pointA)).cross(pointC - pointA);
				normalVector.normalize();

				// for shells, we want shade both triangles (one facing outwards, the other
				// facing inwards), so we will flip the inward normal outwards.

				// to test whether the normal vector points outwards or inwards
				// we only need add a very small vector (in the normal's direction) to
				// one of the vertex and text whether the end point of the vector is inside
				// the sphere or not
				//if (isIn(pointA + normalVector*0.1, sphereCenter, sphereRadius))
					// flip the normalVector
					//normalVector = - normalVector;

				glBegin(GL_TRIANGLES);
				{
					glNormal3f(normalVector[0], normalVector[1], normalVector[2]);
					glVertex3f(pointA[0], pointA[1], pointA[2]);
					glVertex3f(pointB[0], pointB[1], pointB[2]);
					glVertex3f(pointC[0], pointC[1], pointC[2]);
				}glEnd();
			}
		}
	}
}


/**
 * drawTriangles
 *
 * @param vertices vertex geometry location
 * @param triangles triangle incident vertex table
 * @param triangleShellIDs shell ID associated with each triangle
 * @param colorPallet colorPallet for painting the triangles according to their shell ID
 **/
void CMeshCompression::drawTriangles(vector<CPoint3D> vertices, 
									 vector<long> triangles,
									 vector<long> triangleShellIDs,
									 vector<CPoint3D> colorPallet)
{
	CPoint3D pointA, pointB, pointC, pointTemp, color;
	CVector3 normalVector;
	long shellID;
	long i = 0;
	int colorPalletSize = colorPallet.size();
	// CVector3 ls = CVector3(-1.0f, -1.0f, 10.0);

	/*
	color = colorPallet[0];
	glColor3f((GLfloat)color[0], (GLfloat)color[1], (GLfloat)color[2]);
	*/

	for(i= 0; i < (long)triangles.size(); i += 3)
	{
		pointA = vertices[triangles[i]];
		pointB = vertices[triangles[i+1]];
		pointC = vertices[triangles[i+2]];

		if(triangleShellIDs.size() != 0)
		{
			shellID = triangleShellIDs[i/3];

			// if the triangle is associated with an unrecognized shell
			if (-1 == shellID)
			{
				// default to gray
				glColor3f(0.5, 0.5, 0.5);
			}
			else
			{
				// rotate the pallet to select a color
				color = colorPallet[shellID % colorPalletSize];
				glColor3f((GLfloat)color[0], (GLfloat)color[1], (GLfloat)color[2]);
			}
		}
		else
		{
			// default to gray
			glColor3f(0.5, 0.5, 0.5);
		}
		
		normalVector = ((CVector3)(pointB - pointA)).cross(pointC - pointA);
		normalVector.normalize();
		

		// for shells, we want shade both triangles (one facing outwards, the other
		// facing inwards), so we will flip the inward normal outwards.

		// to test whether the normal vector points outwards or inwards
		// we only need add a very small vector (in the normal's direction) to
		// one of the vertex and text whether the end point of the vector is inside
		// the sphere or not
		//if (isIn(pointA + normalVector*0.1, sphereCenter, sphereRadius))
			// flip the normalVector
			//normalVector = - normalVector;

		glBegin(GL_TRIANGLES);
		{
			glNormal3f(normalVector[0], normalVector[1], normalVector[2]);
			glVertex3f(pointA[0], pointA[1], pointA[2]);
			glVertex3f(pointB[0], pointB[1], pointB[2]);
			glVertex3f(pointC[0], pointC[1], pointC[2]);
		}glEnd();
	}
}

void CMeshCompression::drawTrianglesHack(vector<CPoint3D> vertices, vector<long> triangles, CPoint3D color,
											CPoint3D boundingSphereCenter, double boundingSphereRadius)
{
	CPoint3D pointA, pointB, pointC, pointTemp;
	CVector3 normalVector;
	unsigned int iIndex = 0;
	CVector3 ls = CVector3(-1.0f, -1.0f, 10.0);
	glColor3f((GLfloat)color[0], (GLfloat)color[1], (GLfloat)color[2]);

	CPoint3D sphereCenter = CPoint3D(0, 0, 0);
	double sphereRadius = 3;

	for(iIndex = 0; iIndex < triangles.size(); iIndex = iIndex+3)
	{
		pointA = vertices[triangles[iIndex]];
		pointB = vertices[triangles[iIndex+1]];
		pointC = vertices[triangles[iIndex+2]];
		normalVector = ((CVector3)(pointB - pointA)).cross(pointC - pointA);
		normalVector.normalize();
		

		// for shells, we want shade both triangles (one facing outwards, the other
		// facing inwards), so we will flip the inward normal outwards.

		// to test whether the normal vector points outwards or inwards
		// we only need add a very small vector (in the normal's direction) to
		// one of the vertex and text whether the end point of the vector is inside
		// the sphere or not
		if (isIn(pointA + normalVector*0.1, boundingSphereCenter, boundingSphereRadius))
			// flip the normalVector
			normalVector = - normalVector;

		glBegin(GL_TRIANGLES);
		{
			glNormal3f(normalVector[0], normalVector[1], normalVector[2]);
			glVertex3f(pointA[0], pointA[1], pointA[2]);
			glVertex3f(pointB[0], pointB[1], pointB[2]);
			glVertex3f(pointC[0], pointC[1], pointC[2]);
		}glEnd();
	}
}


bool isIn(CPoint3D p, CPoint3D center, double radius)
{
	CVector3 d = p - center;
	if(d.magnitude() < radius)
		return true;
	return false;
}


/**
 * relaxation of points on a shpere
 * O(n^2) algorithm
 *
 * @param radius of influence, outside this radius would be considered
 *		  to have no influence over the current point
 * @param k small constant
 **/
void CMeshCompression::relaxVerticesInBall(vector<CPoint3D>& vertices, CPoint3D ballCenter, 
										   double ballRadius, double radius, double k)
{
	unsigned int i, j;

	CVector3 qp, repulsion, repulsionSum;
	CPoint3D p, p1, p2, q;

	for(i = 0; i < vertices.size(); i++)
	{
		repulsionSum = CVector3(0, 0, 0);
		p = vertices[i];
		for(j = 0; j < vertices.size(); j++)
		{
			q = vertices[j];

			// if q is the same point or q is located 
			// too far away from p, discard the point
			if((p-q).magnitude() > radius || p == q)
				continue;
			
			qp = p - q;
			// the repulsion force is inversely related to the
			// distance.
			repulsion = (p-q).normalized() * (radius/qp.magnitude());
			repulsionSum += repulsion;
		}

		// new position not limited to be inside the ball
		p1 = p + repulsionSum * k;

		// if p1 goes outside the ball, put it back into the ball
		if((p1 - ballCenter).magnitude() > ballRadius) 
		{
			p2 = ballCenter + (p1 - ballCenter).normalized() * ballRadius;
			vertices[i] = p2;
		}
		else
		{
			vertices[i] = p1;
		}		
	}
}



/**
 * ballWithBlocker
 *
 * produce points lies on a ball with blocker
 * 
 * @param nPoints number of points to generate on the given sphere; however, if there is a blocker present
 *        to maintain the same uniform density, we will just remove points already generated, so the
 *        final number of points may be less than this number.
 * @param radiusBlocker the radius of the blocker, if this is set to 0, then there is no blocker
 *
 **/
vector<CPoint3D> CMeshCompression::ballWithBlocker(long nPoints, CPoint3D centerBall, double radiusBall, 
												   CPoint3D centerBlocker, double radiusBlocker)
{
	long i;
	vector<CPoint3D> points;

	double ballVolume = 4 / 3 * PI * radiusBall * radiusBall * radiusBall; 
	double radiusInfluence = 2 * pow((ballVolume / nPoints), 1/3);

	// Seed the random-number generator with current time so that
	// the numbers will be different every time we run.
	srand( (unsigned)time( NULL ) );

	// generate the points
	for(i = 0; i < nPoints; i++)
	{
		CPoint3D p = ballPoint(centerBall, radiusBall);
		points.push_back(p);
	}

	// relax the points
	double k = 0.1;
	for(i = 0; i< 10; i++)
	{
		relaxVerticesInBall(points, centerBall, radiusBall, radiusInfluence, k);
		k -= 0.01;
	}

	// remove the points inside the blocker
	if(radiusBlocker == 0)
	{	
		return points;
	}
	else
	{
		vector<CPoint3D> temp;

		for(i = 0; i < nPoints; i++)
			if (!isIn(points[i], centerBlocker, radiusBlocker))
				temp.push_back(points[i]);
		return temp;
	}
}

/**
 * relaxation of points on a shpere
 * O(n^2) algorithm
 *
 * @param radius of influence, outside this radius would be considered
 *		  to have no influence over the current point
 * @param k small constant
 **/

void CMeshCompression::relaxVerticesOnSphere(vector<CPoint3D>& vertices, CPoint3D sphereCenter, 
											 double sphereRadius, double radius, double k)
{
	unsigned int i, j;

	CVector3 qp, repulsion, repulsionSum;
	CPoint3D p, p1, p2, q;

	for(i = 0; i < vertices.size(); i++)
	{
		repulsionSum = CVector3(0, 0, 0);
		p = vertices[i];
		for(j = 0; j < vertices.size(); j++)
		{
			q = vertices[j];

			// if q is the same point or q is located 
			// too far away from p, discard the point
			if((p-q).magnitude() > radius || p == q)
				continue;
			
			qp = p - q;
			// the repulsion force is inversely related to the
			// distance.
			repulsion = (p-q).normalized() * (radius/qp.magnitude());
			repulsionSum += repulsion;
		}

		// new position not on the sphere
		p1 = p + repulsionSum * k;

		// put p1 back onto the sphere;
		p2 = sphereCenter + (p1 - sphereCenter).normalized() * sphereRadius;
		vertices[i] = p2;
	}
}


/**
 * sphereWithBlocker
 *
 * produce points lies on a sphere with blocker
 * 
 * @param nPoints number of points to generate on the given sphere; however, if there is a blocker present
 *        to maintain the same uniform density, we will just remove points already generated, so the
 *        final number of points may be less than this number.
 * @param radiusBlocker the radius of the blocker, if this is set to 0, then there is no blocker
 *
 **/
vector<CPoint3D> CMeshCompression::sphereWithBlocker(long nPoints, CPoint3D centerSphere, double radiusSphere, 
													CPoint3D centerBlocker, double radiusBlocker)
{
	long i;
	vector<CPoint3D> points;

	// for relaxation of the points on the sphere
	double sphereArea = 4 * PI * radiusSphere * radiusSphere; 
	double radiusInfluence = 2 * sqrt(sphereArea / nPoints);

	// Seed the random-number generator with current time so that
	// the numbers will be different every time we run.
	srand( (unsigned)time( NULL ) );

	// generate the points
	for(i = 0; i < nPoints; i++)
	{
		CPoint3D p = spherePoint(centerSphere, radiusSphere);
		points.push_back(p);
	}

	double k = 0.2;
	// relax the points
	for(i = 0; i< 10; i++)
	{
		relaxVerticesOnSphere(points, centerSphere, radiusSphere, radiusInfluence, k);
		k -= 0.015;
	}

	// remove the points inside the blocker
	if(radiusBlocker == 0)
	{	
		return points;
	}
	else
	{
		vector<CPoint3D> temp;

		for(i = 0; i < nPoints; i++)
			if (!isIn(points[i], centerBlocker, radiusBlocker))
				temp.push_back(points[i]);
		return temp;
	}
}


/* Returns a random point on the sphere with center and radius 
 * Must include stdlib.h for RAND_MAX 
 */
CPoint3D CMeshCompression::spherePoint(CPoint3D center, double radius)
{	
	double x0 = center[0];
	double y0 = center[1];
	double z0 = center[2];
	
	double slice = (((double)rand()/RAND_MAX) - 0.5) * 2 * radius;
	double angle = ((double)rand()/RAND_MAX) * 2 * PI;
	double new_radius = sqrt(radius*radius - slice*slice);
	double new_x = slice + x0;	
	double new_y = new_radius * cos(angle) + y0;
	double new_z = new_radius * sin(angle) + z0;

	return CPoint3D(new_x, new_y, new_z);
}


/* Returns a random point in the ball with center and radius
 * Must include stdlib.h for RAND_MAX
 */
CPoint3D CMeshCompression::ballPoint(CPoint3D center, double radius) 
{
	double x, y, z;
	// random number with [-1, 1]
	do
	{
		x = (((double)rand()/RAND_MAX) - 0.5) * 2;
		y = (((double)rand()/RAND_MAX) - 0.5) * 2;
		z = (((double)rand()/RAND_MAX) - 0.5) * 2;
	}while (x*x + y*y + z*z >= 1);

	CPoint3D pStd = CPoint3D(x, y, z);
	
	return CPoint3D(pStd * radius + center);
}