/** =================================================================\n
 *
 * CS 7491 2004\n
 * @file MeshCompression.h
 *
 * Primary Authors: Brandon Beck
 *					Nguyen Truong (nltruong@cc.gatech.edu)\n
 *                  Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright 2004\n
 * (C) Copyright Georgia Institute of Technology, 2004\n
 *
 * ==================================================================\n
 * triangle mesh compression and simplification task class
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2004-02-13
 * @author Brandon Beck, Nguyen Truong, Howard Zhou
 *
 * @todo
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#ifndef __CMESHCOMPRESSION_H__
#define __CMESHCOMPRESSION_H__

#include <vector>

#include "..\Viewer\stdafx.h"
#include "..\Viewer\Task.h"
#include "..\MatrixLib\Vector.h"

#include "..\Mesh\TriangleMesh.h"

/// @namespace using standard namespace
using namespace std;

// ==================================================================================
/**
 * @class CMeshCompression
 * @brief Mesh Compression Task class
 * @ingroup Mesh
 *
 * @par requirements
 *
 * @par license
 * This code is absolutely free to use and modify. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 *
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2004-02-13
 * @author Brandon Beck, Nguyen Truong, Howard Zhou
 *
 * This class is the mesh compression task class
 * 
 * @todo
 * @bug
 * @warning
 */
// ================================================================================

class CMeshCompression : public CTask
{
private:
	CTriangleMesh _triangleMesh;

	// ------------------------------------------------------
	// private helpers
	// ------------------------------------------------------
	static CPoint3D ballPoint(CPoint3D center, double radius);
	static CPoint3D spherePoint(CPoint3D center, double radius);

public:
	///< static function pointer
	// static void (*renderFunc) (CTask *);

	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------

	/// Default constructor. Creates an uninitialized matrix.
	CMeshCompression():CTask(){};

	// Destructor
	~CMeshCompression(){};

	// ------------------------------------------------------
	// Predicators
	// ------------------------------------------------------

	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------

	// ------------------------------------------------------
	// Overloaded Operators 
	// ------------------------------------------------------

	// ------------------------------------------------------
	// Input, Output, and Test
	// ------------------------------------------------------
	void drawPoints(vector<CPoint3D> points, CPoint3D color);
	void drawTriangleMesh(CTriangleMesh& triangleMesh);

	void drawTriangles(vector<CPoint3D> vertices, vector<long> triangles, 
		vector<long> triangleShellIDs, vector<CPoint3D> colorPallet);
	
							
	// HOWARD: this is a hack until we figure out how to draw two sided triangles
	void drawTrianglesHack(vector<CPoint3D> vertices, vector<long> triangles, CPoint3D color,
						   CPoint3D boundingSphereCenter, double boundingSphereRadius);

	// this renderGLScene will be used as a function pointer to supply the
	// openGL view. Therefore it has to be made static memeber function to be
	// treated as a normal function pointer. Normal memeber functions will be
	// treated as a memeber function pointer because their addresses are also
	// associated with the object they belong.
	static void renderGLScene(CTask * task);

	// ------------------------------------------------------
	// Other operation
	// ------------------------------------------------------
	void run();

	static vector<CPoint3D> ballWithBlocker(long nPoints, CPoint3D centerBall, double radiusBall, 
		CPoint3D centerBlocker = CPoint3D(0, 0, 0), double radiusBlocker = 0);
	static vector<CPoint3D> sphereWithBlocker(long nPoints, CPoint3D centerSphere, double radiusSphere, 
		CPoint3D centerBlocker = CPoint3D(0, 0, 0), double radiusBlocker = 0);

	static void relaxVerticesOnSphere(vector<CPoint3D>& vertices, CPoint3D sphereCenter, 
									  double sphereRadius, double radius, double k);
	static void relaxVerticesInBall  (vector<CPoint3D>& vertices, CPoint3D ballCenter, 
									  double ballRadius, double radius, double k);


	// ------------------------------------------------------
	// Friend functions
	// ------------------------------------------------------
};



#endif // __CMESHCOMPRESSION_H__
